/***************************************************************************
                           facttable.h  - storage for 
			                  facts that occur
					  in modes without outputs
                             -------------------
    begin                : Wed Mar 22 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef FACTTABLE_H
#define FACTTABLE_H
#include "type.h"
#include "farray.h"

class BaseFactTable {
  public:
    BaseFactTable () { ; }
    virtual ~BaseFactTable () { ; }
    virtual bool getFact ( INTTYPE *coor ) = 0;
};

class FactTable:public BaseFactTable {
  public:
    FactTable ( int dimensions );
    virtual ~FactTable ();
    void setFact ( INTTYPE *coor, bool value );
    bool getFact ( INTTYPE *coor ) {
      FArray<unsigned int> *a = ( dimensions > 1 ) ? (*data) [ coor ] : data_onedimension;
      if ( a ) {
        unsigned int co = coor[dimensions - 1];
        unsigned int val = a->element ( co / 32 );
	return ( 1 << ( co % 32 ) ) & val;
      }
      else
        return 0;
    }
    bool existsFact ( INTTYPE *coor, int length ) {
      if ( length == dimensions )
        return getFact ( coor );
      return data->exists ( coor, length );
    }
  private:
    FMultiArray<FArray<unsigned int> > *data;
    FArray<unsigned int> *data_onedimension;
    int dimensions;
};

typedef bool (*TgetFact) ( INTTYPE *coor );
TgetFact extern_lookup_facttable ( char *name );

class ExternalFactTable:public BaseFactTable {
  public:
    ExternalFactTable ( char *functionname ) {
      function = extern_lookup_facttable ( functionname );
    }
    bool getFact ( INTTYPE *coor ) {
      return function ( coor );
    }
  private:
    TgetFact function;
};

#endif

