/***************************************************************************
                          modefacttable.h  - storage for facts
			                     that occur in modes
					     with outputs
                             -------------------
    begin                : Tue Apr 11 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef MODEFACTTABLE_H
#define MODEFACTTABLE_H
#include "typetable.h"
#include "type.h"

struct ListModeFactCoorsElement {
  INTTYPE *value;
  ListModeFactCoorsElement *next;
};

class ListModeFactCoorsStore {
  public:
    ListModeFactCoorsElement *first;
    ListModeFactCoorsStore ();
    ~ListModeFactCoorsStore ();
};

#include "extern.h"
#include "farray.h"

struct TypeTable;

class BaseModeFactTable {
  public:
    BaseModeFactTable () { ; }
    virtual ~BaseModeFactTable () { ; }
    virtual ListModeFactCoorsElement *getListFacts ( INTTYPE *coor ) = 0;
    virtual bool existsFact ( INTTYPE *coor, int length ) = 0;
};

class TypeTable;
struct ModeData;

/**The ModeFactTable allows fast lookup for a given mode to a fact table by saving search results.
  */
class ModeFactTable:public BaseModeFactTable {
  public:
    ModeFactTable ( TypeTable &typetable, ModeData *modedata  );
    virtual ~ModeFactTable();
    inline ListModeFactCoorsElement* getListFacts ( INTTYPE *coor ) {
      ListModeFactCoorsStore *el = data->element ( coor );
      return el ? el->first : NULL;
    }
    inline bool existsFact ( INTTYPE *coor, int length ) {
      return data->exists ( coor, length );
    }
    
     // combined coordinates and value
    void setFact ( INTTYPE *coor );
    
    double meanListLength () { return nroflists ? sumoflistlengths / nroflists : 0; }
  private:
    int nroflists;
    int sumoflistlengths;
    FMultiArray<ListModeFactCoorsStore> *data;
    ModeData *modedata;
    int inputdim, outputdim;
};

typedef ListModeFactCoorsElement* (*TGetListFacts) ( INTTYPE *coor );
TGetListFacts extern_lookup_modefacttable ( char *name );

class ExternalModeFactTable:public BaseModeFactTable {
  public:
    ExternalModeFactTable ( char *functionname ) {
      function = extern_lookup_modefacttable ( functionname );
    }
    ListModeFactCoorsElement* getListFacts ( INTTYPE *coor ) {
      return function ( coor );
    }
    bool existsFact ( INTTYPE *coor, int length ) {
      return true;
    }
  private:
    TGetListFacts function;
};

#include "modetable.h"
#endif
