/***************************************************************************
                          modetable.cpp  -  the dictionary of modes
                             -------------------
    begin                : Wed Mar 22 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "modetable.h"
#include "modefacttable.h"
#include "predicatetable.h"
#include "globaldata.h"
#include "type.h"
#include "output.h"
#include <string.h>
#include <values.h>

ModeTable::ModeTable ():
  filledarray ( false ),
  _size ( 0 ), cdepsize ( 0 ), variablevsconstantcheck ( false ) {

}

ModeTable::~ModeTable () {
  for ( int i = 0; i < deeparray.size (); i++ )
    if ( deeparray[i] != NULL ) {
      //delete deeparray[i]->params;
      delete deeparray[i];
    }
  for ( int i = 0; i < _size; i++ )
    delete array[i].params;
}

// params: the constructed mode parameters.
// mpd: the original mode parameters.
void ModeTable::fillAllModes ( ModeParamData *params, ModeData *md, int &fillsize, int position, ModeParamData *mpd ) {
  if ( position < md->size )
    if ( IS_CONST_PARAM ( mpd + position ) ) {
      TypeData *td = typetable.typeData ( mpd[position].number );
      for ( int i = 0; i < td->constnumbers; i++ ) {
        if ( !td->forbidden.element ( i ) ) {
  	  params[position].number = i;
          params[position].type = Constant;
          fillAllModes ( params, md, fillsize, position + 1, mpd );
	}
      }
    }
    else {
      params[position] = mpd[position];
      fillAllModes ( params, md, fillsize, position + 1, mpd );
    }
  else {
    array[fillsize].nr = fillsize;
    array[fillsize].data = md;
    array[fillsize].params = new ModeParamData[md->size];
    for ( int i = 0; i < md->size; i++ )
      array[fillsize].params[i] = params[i];
    fillsize++;
  }
}

void ModeTable::createArray () {
  int i, j;
  _size = 0;
  int currentmodenr;

  // we first compute the size of the array.
  for ( i = 0; i < deeparray.size (); i++ ) {
    currentmodenr = 1;
    for ( j = 0; j < deeparray[i]->size; j++ ) {
      if ( IS_CONST_PARAM ( deeparray[i]->params + j ) )
        currentmodenr *= typetable.typeData ( deeparray[i]->params[j].number )->allowedNrOfConstants ();
    }
    deeparray[i]->nrofmodes = currentmodenr;
    deeparray[i]->activefrequent.setActive ( currentmodenr );
    _size += currentmodenr;
  }

  array = new ModeConstData[_size];
  int fillsize = 0;

  for ( i = 0; i < deeparray.size (); i++ ) {
    ModeParamData params[deeparray[i]->size];
    deeparray[i]->firstnr = fillsize;
    fillAllModes ( params, deeparray[i], fillsize, 0, deeparray[i]->params );
  }
}

void ModeTable::fillMinChainLength ( ModeData *fillmd, ModeData *walkmd, int length ) {
  if ( !walkmd->haschainsucc ) {
    if ( length < fillmd->minchainlength )
      fillmd->minchainlength = length;
  }
  else {
    for ( int i = 0; i < walkmd->chainedmodes.size (); i++ )
      fillMinChainLength ( fillmd, walkmd->chainedmodes[i], length + 1 );
  }
}


void ModeTable::determineQuickPreEval ( ModeData *md ) {
  int keytype = key->params[0].number;

  if ( ( md->params[0].type == Plus ||
         md->params[0].type == PlusEx ||
         md->params[0].type == ChainPlus ) &&
	 md->params[0].number == keytype ) {
    int i = 1;
    while ( i < md->size &&
            ( md->params[i].type == Constant || md->params[i].type == ChainConstant ) )
      i++;
    md->quickpreevalsize = i;
    if ( i == md->size ) {
      md->quickpreeval = true;
      return;
    }
    while ( i < md->size &&
            md->params[i].type != Constant && md->params[i].type != ChainConstant )
      i++;
    if ( i == md->size )
      md->quickpreeval = true;
    else
      md->quickpreeval = false;
  }
  else
    md->quickpreeval = false;
}

void ModeTable::determinePKeys ( ModeData *md, PredicateData *pd ) {
  if ( pd->primarykeys ) {
    FListWalk<FArray<bool> *> walk ( pd->primarykeys );
    FArray<bool> **current;
    FArray<int> *pkeyparams;
    
    while ( current = walk.current () ) {
      if ( pkeyparams = isPKeySubSetOfMode ( *current, md ) ) {
        if ( !md->primarykeys )
	  md->primarykeys = new FList<FArray<int> *>;
	md->primarykeys->append ( pkeyparams );
      }
      walk.next ();
    }
  }
}

void ModeTable::fillDependencies () {
  bool modepattern[deeparray.size ()], typepattern[typetable.size ()];
  int i, j, k, nr, *params;
  PredicateData *pd;
  ModeData *md;
  bool equivalent, equal;

  createArray ();

  for ( i = 0; i < deeparray.size (); i++ ) {
    md = deeparray[i];
    pd = predicatetable.predicateData ( md->predicate );
    
    determinePKeys ( md, pd );
    
    determineQuickPreEval ( md );
    
    fillMinChainLength ( md, md, 0 );

    if ( md->outputs ) {
      for ( j = 0; j < typetable.size (); j++ )
	typepattern[j] = false;
      for ( j = 0; j < deeparray[i]->size; j++ )
	if ( md->params[j].type == Min || md->params[j].type == MinEx )
          typepattern[md->params[j].number] = true;
    }
    nr = 0;

    for ( j = 0; j < deeparray.size (); j++ ) {
      if ( md->outputs ) {	
	modepattern[j] = false;
	for ( k = 0; k < deeparray[j]->size; k++ ) {
          if ( !deeparray[j]->haschainpred && 
	       ( deeparray[j]->params[k].type == Plus || deeparray[j]->params[k].type == PlusEx ) && 
	       typepattern[deeparray[j]->params[k].number] ) {
            modepattern[j] = true;
            nr++;
            break;
          }
	}
      }
      if ( deeparray[j]->predicate == md->predicate && i != j ) {
        // Here the user's specificatons are checked.
	// No two modes may exist which are equal.
	// The check is not straightforward because of ChainedPlus and
	// ChainedConstant which are treated as Plus and Constant in this
	// respect.
	// If two modes are equal, but are both not the first in a
	// chain, they are both allowed. The reason?
	// Given a query with a certain order, we should be able to determine
	// for each atom deterministically which mode introduced this atom.
	
	// This reasoning is too short-cutted. We need this feature for the
	// mutagenesis dataset.
	// We should relax it.
	// equal modes between first in chain and later in chain
	// should be allowed. In any query, we can determine which one
	// introduced an atom as chained modes have preference always.
	// The equal atom generated by the other mode is generated
	// later and should be pruned by the equivalency mechanism.
	// Therefore, it is required that the equivalent mode is put into
	// the equivalency list.
	
	// The start of the chain deterministically determines the modes
	// that follow.
	// Some issues have not been resolved cleanly.
	// Assume: - mode(1,p(+,-1)). mode(c,q(+1,#)).
	//         - mode(1,q(+,#)).
	//
	if ( !deeparray[i]->haschainpred && !deeparray[j]->haschainpred ) {
   	  equal = true;
	  for ( k = 0; k < deeparray[j]->size && equal; k++ )
	    switch ( deeparray[j]->params[k].type ) {
	      case Plus: case ChainPlus: case PlusEx:
	        if ( deeparray[i]->params[k].type != Plus &&
	             deeparray[i]->params[k].type != ChainPlus &&
		     deeparray[i]->params[k].type != PlusEx )
		  equal = false;
		break;
	      case MinEx:
	        if ( deeparray[i]->params[k].type != MinEx )
		  equal = false;
		break;
	      case Constant: case ChainConstant:
	        if ( deeparray[i]->params[k].type != Constant &&
		     deeparray[i]->params[k].type != ChainConstant )
		  equal = false;
	  }
	  if ( equal ) {
	    cerr << "Illegal bias specification --- mode "
	         << deeparray[i]->deepnr << " and " << deeparray[j]->deepnr
		 << " are equal!" << endl;
            exit ( 1 );		 
	  }
	}
	
	equivalent = true;
	for ( k = 0; k < deeparray[j]->size; k++ ) {
	  if ( ( md->params[k].type == Plus || md->params[k].type == MinEx ||
	         md->params[k].type == ChainPlus || md->params[k].type == PlusEx )
		 && ( deeparray[j]->params[k].type == Constant ||
		      deeparray[j]->params[k].type == ChainConstant ) ) {
	    equivalent = false;
	    break;
	  }
	  else
    	    if ( ( deeparray[j]->params[k].type == Plus || deeparray[j]->params[k].type == MinEx ||
	           deeparray[j]->params[k].type == ChainPlus || deeparray[j]->params[k].type == PlusEx )
		   && ( md->params[k].type == Constant ||
		        md->params[k].type == ChainConstant ) ) {
	      equivalent = false;
	      break;
	    }
	}
	if ( equivalent ) {
	  md->equivalentmodes.append ( deeparray[j]->firstnr );
	  //if ( !deeparray[j]->outputs ) {
	    // there reverse is also equivalent, but we don't have a separate loop
	    // for filling the reverse case.
	  //  deeparray[j]->equivalentmodes.append ( md->firstnr );
	  //}
	}
	else {
	  // two modes of the same predicate are not equivalent: in one there is a constant
	  // at a position, while the other has a variable. We mark this, as it requires
	  // more checks during evaluation.
	  variablevsconstantcheck = true;
	}
      }
    }
    
    if ( md->haschainpred ) {
      // usually, the same mode cannot generate exactly the same atom twice.
      // Atoms in a chain may however be generated after several first atoms.
      // This is certainly not valid for all kinds of series, but for the
      // moment the candidate generation will have to make sure that no
      // duplicates are generated.
      
      // Actually, if a chained atom refers to at least one variable
      // which was introduced by a previous atom in the chain, then this
      // atom could not have been introduced earlier.
      // Otherwise, so when there are only:
      // - normal pluses
      // - and/or constants, either normal or chained
      // this atom could have been introduced earlier.
      
      bool useschainedplusvariable = false;
      for ( int i = 0; i < md->size; i++ )
        if ( md->params[i].type == ChainPlus ) {
	  ModeData *md2 = chaingoback ( md, -md->params[i].back );
	  if ( md2->params[md->params[i].backparamnr].type == MinEx ) {
	    useschainedplusvariable = true;
	    break;
	  }
	}
      if ( !useschainedplusvariable ) {
        md->equivalentmodes.append ( md->firstnr );
      }
    }
    
    int nr2;
    
    if ( md->haschainsucc ) {
      // can modes without outputs have equivalent modes?
      // (Where equivalent means NOT EQUAL)
      // Yes! m(+,1,#) (where 1 is constant chain), and m(+,#,#).
      md->depmodsize = nr + md->chainedmodes.size ();
      md->dependentmodes = new int[nr+md->chainedmodes.size ()];
      for ( int j = 0; j < md->chainedmodes.size (); j++ )
        md->dependentmodes[j] = md->chainedmodes[j]->firstnr;
      nr2 = md->chainedmodes.size ();
    }
    else {
      md->depmodsize = nr;
      md->dependentmodes = new int[nr];
      nr2 = 0;
    }
    if ( nr > 0 )
      for ( j = 0; j < deeparray.size (); j++ )
        if ( modepattern[j] )
          md->dependentmodes[nr2++] = deeparray[j]->firstnr;
  }

  filledarray = true;
  
  checkNOOI ();
}

void ModeTable::createTableForMode ( ModeData *md ) {
  PredicateData *pd = predicatetable.predicateData ( md->predicate );
  md->reuses = false;
  md->equalingmode = false;
  int i = md->deepnr, j, k;

  if ( md->alwaystrue )
    md->table.facttable = NULL;
  else
  
  if ( md->outputs )
    if ( pd->key )
      md->table.modefacttable = NULL; // no facttable, this is the key
    else {
       // search for a previous mode which might have another constant
       // as `input' or another output mode, but is the same otherwise,
       // and reuse the table associated with that mode.
      BaseModeFactTable *mft = NULL;

      for ( j = 0; j < i; j++ )
        if ( deeparray[j]->predicate == deeparray[i]->predicate ) {
	  // check parameters
	  for ( k = 0; k < deeparray[i]->size; k++ )
	    if (    ( IS_OUTPUT_PARAM ( deeparray[j]->params + k ) && IS_INPUT_PARAM ( deeparray[i]->params + k ) )
	         || ( IS_INPUT_PARAM ( deeparray[j]->params + k ) && IS_OUTPUT_PARAM ( deeparray[i]->params + k ) ) )
	      break;
	  if ( k == deeparray[j]->size ) {
	    mft = deeparray[j]->table.modefacttable;
	    break;
	  }
	}
      if ( !mft )
	if ( md->functionname )
	  md->table.modefacttable = new ExternalModeFactTable ( md->functionname );
	else
          md->table.modefacttable = new ModeFactTable ( typetable, md );
      else {
	md->reuses = true;
	md->table.modefacttable = mft;
      }
    }
  else {
    md->reuses = false;
    md->equalingmode = false;
    for ( j = 0; j < deeparray.size (); j++ )
      if ( i != j && deeparray[j]->predicate == md->predicate )
        for ( k = 0; k < deeparray[i]->size; k++ ) {
          if ( ( md->params[k].type == Plus || md->params[k].type == ChainPlus || md->params[k].type == PlusEx ) &&
               IS_OUTPUT_PARAM ( deeparray[j]->params + k ) ) {
            md->equalingmode = true;
            goto quitloops;
          }
        }
quitloops:
    if ( !pd->facttable )
      if ( md->functionname )
	pd->facttable = new ExternalFactTable ( md->functionname );
      else
        pd->facttable = new FactTable ( md->size );
    else
      md->reuses = true;
    md->table.facttable = pd->facttable;
  }

  if ( !md->reuses )
    pd->modes.append ( md );
}

// Prepend the current mode to the deeptable.
// The numbers of all existing modes must be addapted as far as filled in.
// This function makes sure that chained modes are the first modes in any child list.
void ModeTable::prepend ( ModeData *md ) {
  cdepsize++;
  for ( int k = deeparray.size (); k > 0; k-- ) {
    ModeData *md2 = deeparray [ k - 1 ];
    md2->deepnr++;
    md2->chainprednr++;
    if ( md2->chaindeepnr != -1 )
      md2->chaindeepnr++;
    deeparray.insertAt ( k, md2 );
  }
  md->deepnr = 0;
  md->chainprednr++;
  if ( md->chaindeepnr != -1 )
    md->chaindeepnr++;
  deeparray.insertAt ( 0, md );
}

ModeData* ModeTable::addMode ( int predicate, int parsize, ModeParamData *params, char *functionname, ModeData *mdchainpred ) {
  bool outputs = 0;
  int i, j, k, nrinput = 0, nroutput = 0, nrconst = 0, nrplusex = 0;

  ModeData *data;
  data = new ModeData;
  data->deepnr = deeparray.size ();
  data->predicate = predicate;
  data->size = parsize;
  data->params = params;
  data->haschainsucc = false;
  data->minchainlength = MAXINT;
  data->chaindeepnr = -1;
  if ( mdchainpred ) {
    data->haschainpred = true;
    data->chainprednr = mdchainpred->deepnr;
    deeparray[data->chainprednr]->haschainsucc = true;
    deeparray[data->chainprednr]->chaindeepnr = data->chainprednr;
    // prepend this mode!!!
    FArray<ModeData*> &a = deeparray[data->chainprednr]->chainedmodes;
    a.append ( data ); // allocate space
    for ( int i = 0; i < a.size () - 1; i++ )
      a[i+1]=a[i];
    a[0] = data;
  }
  else {
    data->haschainpred = false;
  }
  data->haschainconst = false;
  init ( data->pattern );
  if ( functionname ) {
    data->functionname = new char[strlen ( functionname ) + 1];
    strcpy ( data->functionname, functionname );
  }
  else
    data->functionname = NULL;

  for ( i = 0; i < parsize; i++ )
    if ( params[i].type == Plus || params[i].type == PlusEx || params[i].type == ChainPlus ) {
      if ( params[i].type == PlusEx )
        nrplusex++;
      nrinput++;
      setOn ( data->pattern, i );
    }
    else
      if ( params[i].type == MinEx ) {
        nroutput++;
        outputs = 1;
      }
      else
        if ( params[i].type == Constant ) 
	  nrconst++;
	else
	  if ( params[i].type == ChainConstant ) {
	    nrconst++;
	    data->haschainconst = true;
	  }

  data->outputs = outputs;
  data->nrinput = nrinput;
  data->nrconst = nrconst;
  data->nrplusex = nrplusex;
  data->inputvarpos = new int[nrinput];
  data->nroutput = nroutput;
  data->outputvarpos = new int[nroutput];

  for ( i = 0, j = 0, k = 0; i < parsize; i++ )
    if ( params[i].type == Plus || params[i].type == PlusEx || params[i].type == ChainPlus )
      data->inputvarpos[j++] = i;
    else
      if ( params[i].type == MinEx )
        data->outputvarpos[k++] = i;

  data->dependentmodes = NULL; // only filled when building an array
  data->depmodsize = 0;

  if ( mdchainpred )
    prepend ( data );
  else
    deeparray.append ( data );

  createTableForMode ( data );

  return data;
}

void ModeTable::print () {
  for ( int i = 0; i < size (); i++ )
    modeData ( i );
  return;
  int i, j, w;

  if ( filledarray ) {
    cout << "+---+-----------+-----------------+--------------------+" << endl
         << "| N | predicate | params          | dependent modes    |" << endl
         << "+---+-----------+-----------------+--------------------+" << endl;
    for ( i = 0; i < size (); i++ ) {
      w = 0;
      ModeData *md = modeData ( i );
      cout << "|" << setw ( 2 ) << md->nr << " |" << setw ( 10 ) << md->predicate << " | ";
      for ( j = 0; j < md->size; j++ ) {
        switch ( md->params[j].type ) {
          case Plus: cout << "+"; break;
	  case PlusEx: cout << "="; break;
	  case ChainPlus: cout << "+(" << md->params[j].back << "," << md->params[j].backparamnr << ")"; break;
	  case Min: cout << "-"; break;
	  case MinEx: cout << "\\"; break;
	  case Constant: cout << "#"; break;
	  case ChainConstant: cout << "#(" << md->params[j].back << "," << md->params[j].backparamnr << ")"; break;
	}
	cout << setw ( 2 ) << md->params[j].number;
	w += 3;
	if ( j < md->size - 1 ) {
	  cout << ", ";
	  w += 2;
	}
      }
      cout << setw ( 18 - w ) << " | ";
      w = 0;
      for ( j = 0; j < md->depmodsize; j++ ) {
        cout << setw ( 2 ) << md->dependentmodes[j];
        w += 2;
        if ( j < md->depmodsize - 1 ) {
          cout << ", ";
          w += 2;
        }
      }
      cout << setw ( 20 - w ) << "|";
      for ( j = 0; j < md->equivalentmodes.size (); j++ )
        cout << "E" << md->equivalentmodes[j] << " ";
      if ( md->haschainpred )
        cout << "<";
      if ( md->haschainsucc )
        cout << ">";
      for ( j = 0; j < md->chainedmodes.size (); j++ )
        cout << "C" << md->chainedmodes[j]->firstnr << " ";
      if ( md->minchainlength != MAXINT )
        cout << "L" << md->minchainlength << " ";
      cout <<  "P" << md->pattern << endl;
    }
    cout << "+---+---+---+-----------+-----------------+--------------------+" << endl;
  }
  else
    cerr << "Error: display not possible for unfilled data." << endl;
}

void setFact ( ModeData *md, INTTYPE *coors ) {
  if ( !md->reuses )
    if ( !md->functionname )
      if ( md->outputs )
        ((ModeFactTable*) md->table.modefacttable)->setFact ( coors );
      else
        ((FactTable*) md->table.facttable)->setFact ( coors, true );
}

// returns the mode which has the input/output pattern as required by
// m. The returned mode never has a predecessor in a chain.
// The computation is more complex in case modenr is a part of a chain
// and contains a constant parameter.
int ModeTable::searchEquivalentMode ( Node *node, ModePatternType m ) {
  int modenr = MODE ( node );
  ModeData *md = array[modenr].data; // no need to use modeData ().
  FArray<int> &equivalent = md->equivalentmodes;

  if ( !md->haschainpred && m == md->pattern ) {
    return modenr;
  }
    
  // for the moment, we perform a linear search. This could be more efficient.
  for ( int i = 0; i < equivalent.size (); i++ ) {
    ModeData *md2 = array[equivalent[i]].data;
    if ( !md2->haschainpred && md2->pattern == m ) {
      if ( md->haschainconst ) {
        // This case checks whether a successor in a chain can be mapped
	// to a first in a chain. Even in the bias restricted to
	// one mode for each input/output type, this is necessary.
	// Look at:
	// mode(c,t(+,-,#c)). (#c:chained constant)
	// mode(1,t(-,+,#)). (#: no chained constant)
	// These two modes are allowed, and still the first has
	// to be translated into the second.
	int diff = 0;
	PredicateData *pd = predicatetable.predicateData ( md->predicate );
	int *pparams = pd->params;
	INTTYPE *nparams = node->params;
	ModeParamData *mpd = array[modenr].params;
	// we know the constant parameters, but we cannot use the mode number
	// to compute the mode number of a mode in which all parameters are constant.
	// We use the generation order here.
	for ( int i = 0; i < md->size; i++, mpd++, pparams++, nparams++ )
	  if ( mpd->type == ChainConstant || mpd->type == Constant ) {
	    TypeData *td = typetable.typeData ( *pparams );
            diff = diff * td->allowedNrOfConstants () + *nparams - td->nrforbidden;
	  }
	return md2->firstnr + diff;
      }
      else
	return md2->firstnr + ( modenr - md->firstnr );
    }
  }
  return -1;
}

void ModeTable::inactivateAllEquivalent ( Node *node ) {
  ModeData *md = array[node->mode].data;
  FArray<int> &equivalentmodes = md->equivalentmodes;
  if ( md->haschainconst )
    return; // for the moment, not supported.
  
  for ( int i = 0; i < equivalentmodes.size (); i++ ) {
    int fmodenr = equivalentmodes[i];
    ModeData *md2 = array[fmodenr].data;
    if ( !md2->haschainconst ) {
      int omodenr = node->mode - md->firstnr;
      md2->activefrequent.setInActive ( omodenr );
    }
  }
}

void ModeTable::checkNOOI () {
  int i;
  
  for ( i = 0; i < deepsize (); i++ )
    if ( !checkNOOIMode ( i ) )
      stopFarmer ();
}

bool ModeTable::checkNOOIMode ( int modenr ) {
  ModeData *md = deeparray[modenr];
  PredicateData *pd = predicatetable.predicateData ( md->predicate );
  int i, j;
  
  if ( pd->primarykeys ) {
  
    FListWalk<FArray<bool> *> walk ( pd->primarykeys );
    FArray<bool> **current;
    
    bool typeimpoi[typetable.size ()];
    for ( i = 0; i < typetable.size (); i++ )
      typeimpoi[i] = false;
    for ( i = 0; i < md->nroutput; i++ ) {
      int typenr = pd->params[ md->outputvarpos[i] ];
      typeimpoi[typenr] = true;
    }

    // check each primary key with each output parameter.
    while ( current = walk.current () ) {
      // check whether one of the outputs is part of a key and should
      // be evaluated without object identity.
      // In that case, this key cannot be used to allow OI.
      
      for ( i = 0; i < md->nroutput; i++ ) {
	int typenr = pd->params[ md->outputvarpos[i] ];
	if ( !typetable.typeData ( typenr )->oi && 
	     (**current) [ md->outputvarpos[i] ] )
	  break;
      }
      if ( i == md->nroutput ) {
        for ( i = 0; i < md->nroutput; i++ ) {
	  int typenr = pd->params[ md->outputvarpos[i] ];
	  if ( !typetable.typeData ( typenr )->oi &&
	       ! (**current) [ md->outputvarpos[i] ] )
	    typeimpoi[typenr] = false;
	}
      }
      walk.next ();
    }
    for ( i = 0; i < typetable.size (); i++ )
      if ( typeimpoi[i] && !typetable.typeData ( i )->oi ) {
	ERR << "Type " << typetable.getTypeName ( i ) 
	    << " cannot be evaluated without OI: " << endl
	    << "predicate " << pd->name << " has this type as output while it is part of a key; bias error." << endl;
        ERR_F
        return false;
      }
  }
  else {
    for ( i = 0; i < md->nroutput; i++ ) {
      int typenr = pd->params[ md->outputvarpos[i] ];
      if ( !typetable.typeData ( typenr )->oi ) {
        ERR << "Type " << typetable.getTypeName ( typenr ) 
	    << " cannot be evaluated without OI: " << endl
	    << "predicate " << pd->name << " has no keys, and this type is an output for a mode; bias error." << endl;
        ERR_F
        return false;
      }
    }
  }
  
  return true;
}
