/***************************************************************************
                          nametable.cc  - implements a look-up table
                             -------------------
    begin                : Tue Mar 21 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "nametable.h"
#include <string.h>

int hash ( char *s ) {
  char *p;
  unsigned h = 0, g;
  
  for ( p = s; *p != '\0'; p++ ) {
    h = ( h << 4 ) + (*p);
    g = h & 0xf0000000;
    if ( g ) {
      h = h ^ ( g >> 24 );
      h = h ^ g;
    }
  }
  
  return h % TABLESIZE;
}

NameTable::NameTable () {
  for ( int i = 0; i < TABLESIZE; i++ )
    array[i] = NULL;
  nrofnames = 0;
  allnamesfirst = allnameslast = NULL;
}

NameTable::~NameTable () {
 TableListElt *walk, *next; 

 for ( int i = 0; i < TABLESIZE; i++ ) {
   walk = array[i];
   
   while ( walk ) {
     next = walk->next;
     delete walk->name;
     delete walk;
     walk = next;
   }
 }
}

int NameTable::add ( char *name, void *data ) {
  int nr = hash ( name );
  TableListElt *newelt = new TableListElt;
  newelt->name =strdup ( name );
  newelt->data = data;
  newelt->next = array[nr];
  newelt->ordernumber = nrofnames;
  newelt->tnext = NULL;
  array[nr] = newelt;

  if ( !allnamesfirst )
    allnamesfirst = allnameslast = newelt;
  else {
    allnameslast->tnext = newelt;
    allnameslast = newelt;
  }
  nrofnames++;

  return newelt->ordernumber;
}

void *NameTable::lookup ( char *name, int *ordernumber ) {
  TableListElt *walk = array[hash ( name )];

  while ( walk && strcmp ( walk->name, name ) )
    walk = walk->next;

  if ( walk ) {
    if ( ordernumber )
      *ordernumber = walk->ordernumber;
    return walk->data;
  }
  else {
    if ( ordernumber )
      *ordernumber = -1;
    return NULL;
  }
}

int NameTable::ordernumber ( char *name ) {
  TableListElt *walk = array[hash ( name )];
  
  while ( walk && strcmp ( walk->name, name ) )
    walk = walk->next;
  
  if ( walk ) 
    return walk->ordernumber;
  else
    return -1;
}

char *NameTable::lookup ( int n, void **fdata ) {
  int size = n;
  TableListElt *walk = allnamesfirst;
  for ( int i = 0; i < size; i++ )
    walk = walk->tnext;
  if ( fdata )
    *fdata = walk->data;
  return walk->name;
}
