/***************************************************************************
                          parser.cpp    -  implements a parser pf
			                   Prolog facts including functions
                             -------------------
    begin                : Tue Mar 21 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "parser.h"
#include "output.h"
#include <string.h>
#define BUFSIZE 300000

char *buffer;
AtomString_callback callback;

char *strndup ( char *string, int n ) {
  char *ret = new char[n+1];
  int i;
  
  for ( i = 0; i < n; i++ )
    ret[i] = string[i];
  ret[i] = '\0';

  return ret;
}

void readspaces ( char *line, int &pos ) {
  while ( line[pos] == ' ' )
    pos++;
}

AtomString *readpredicate ( char *line, int &pos, char *lastchar = NULL );

int readparams ( char *line, int &pos, AtomString *atom, int actunr ) {
  AtomString *paramatom;
  char lastchar;

  paramatom = readpredicate ( line, pos, &lastchar );
  if ( paramatom ) {
    if ( lastchar == ',' ) {
      pos++;
      actunr = readparams ( line, pos, atom, actunr + 1 );
      if ( actunr == -1 )
        return -1;
      atom->params[actunr] = paramatom;
      if ( actunr < 1 )
        actunr = 1;
      return actunr - 1;
    }
    else
      if ( lastchar == ')' ) {
        pos++;
	atom->nrofparams = actunr + 1;
	atom->params = new AtomString*[actunr+1];
	atom->params[actunr] = paramatom;
	if ( actunr < 1 )
	  actunr = 1;
	return actunr - 1;
      }
      else
	return -1;
  }
  else
    return -1;
}

AtomString *readpredicate ( char *line, int &pos, char *lastchar ) {
  int n = 1;
  AtomString *atom;
  char *old;
  readspaces ( line, pos );
  
    // predicate?
  if ( line[pos] == ',' || line[pos] == '(' || line[pos] == ')' || line[pos] > 'z' || line[pos] < '!' )
    return NULL;
  atom = new AtomString;
  while ( ! ( line[pos + n] == ',' || line[pos + n] == '(' || line[pos + n] == ')' || line[pos + n] > 'z' || line[pos + n] < '!' ) )
    n++;
    // don't allocate new string
  atom->name = line + pos;
  atom->nrofparams = 0;
  pos += n;
  old = line + pos;
  readspaces ( line, pos );
  if ( line[pos] == '(' ) {
    pos++;
    if ( readparams ( line, pos, atom, 0 ) == -1 )
      return NULL;
    readspaces ( line, pos );
  }
  if ( lastchar )
    *lastchar = line[pos];
  *old = '\0';
  return atom;
}

bool parseline ( char *line ) {
  AtomString *atom;
  int pos = 0;
  bool ret;

  readspaces ( line, pos );
    // empty line?
  if ( line[pos] == '\0' )
    return true;
    // comment?
  if ( line[pos] == '/' )
    if ( line[pos + 1] == '/' || line[pos + 1] == '*' ) 
        // allow // and /* to begin comment. NOTE: for one line only,
        // thus no */ recognized 
      return true;
    else
      return false;
      
    // predicate?
  atom = readpredicate ( line, pos );
  
  if ( atom ) {
    ret = callback ( atom );
    free_atomstring ( atom );
    return ret;
  }
  else
    return false;
}

bool parseprolog ( ifstream &input ) {
  bool retval = true;
  int cnt = 1;
  while ( !input.eof () && retval ) {
    if ( !parseline ( buffer ) ) {
      retval = false;
      ERR << "error in line " << cnt << "." << endl;
      ERR_F
    }
    cnt++;
    input.getline ( buffer, BUFSIZE );
  }

  return retval;
}

bool parsefile ( const char *filename, char flag );

bool parsefilelist ( ifstream &input, char flag ) {
  bool retval = true;
  char *lastspace;

  input.getline ( buffer, BUFSIZE );
  while ( !input.eof () ) {
    lastspace = strrchr ( buffer, ' ' );
    *lastspace = '\0';
    do
      lastspace++;
    while ( *lastspace != '\0' && flag != *lastspace );
    if ( *lastspace != '\0' ) {
      ERR << "[ Reading " << buffer << " ]" << endl;
      ERR_F
      retval = parsefile ( buffer, flag );
      if ( !retval ) {ERR << "Aborting" << endl;
        break;}
    }
    input.getline ( buffer, BUFSIZE );
  }

  return retval;
}

bool parsefile ( const char *filename, char flag ) {
  ifstream input;
  bool retval = true;
  input.open ( filename );

  input.getline ( buffer, BUFSIZE );

  // look at signature to determine how to read the file
  if ( !strcmp ( buffer, "FILELIST" ) )
    retval = parsefilelist ( input, flag );
  else
    retval = parseprolog ( input );

  input.close ();

  return retval;
}

bool parse ( const char *filename, AtomString_callback _callback, char flag ) {
  bool retval;

  buffer = new char[BUFSIZE];
  callback = _callback;

  retval = parsefile ( filename, flag );

  delete[] buffer;

  return retval;
}

void printspaces ( int nr ) {
  for ( int i = 0; i < nr; i++ )
    cout << " ";
}

void print_atomstring_recursive ( AtomString *atom, int depth ) {
  printspaces ( depth );
  cout << atom->name;
  if ( atom->nrofparams > 0 ) {
    cout << " (" << endl;
    for ( int i = 0; i < atom->nrofparams; i++ )
      print_atomstring_recursive ( atom->params[i], depth + 2 );
    printspaces ( depth );
    cout << ")";
  }
  cout << endl;
}

bool print_atomstring ( AtomString *atom ) {
  print_atomstring_recursive ( atom, 0 );
  return true;
}

void free_atomstring ( AtomString *atom ) {
  if ( atom->nrofparams ) {
    for ( int i = 0; i < atom->nrofparams; i++ )
      if ( atom->params[i] )
	free_atomstring ( atom->params[i] );
    delete[] atom->params;
  }
  delete atom;
}
