/***************************************************************************
                          treegenerate.cpp  - generates candidate queries,
			                      uses treecount.cpp to count 
					      them, and prints the 
					      frequent queries
                             -------------------
    begin                : Sat Mar 25 2003
    copyright            : (C) 2003 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <string>
#include <iomanip>
#include "globaldata.h"
#include "modetable.h"
#include "typetable.h"
#include "tree.h"
#include "output.h"

using namespace std;

TidData::~TidData () {
  if ( assignments )
    delete assignments;
}

void printNode ( Node *node ) {
  ModeData *md = modetable.modeData ( MODE ( node ) );
  PredicateData *pd = predicatetable.predicateData ( md->predicate );
  cout << pd->name << "(";
  for ( int i = 0; i < md->size; i++ ) {
    switch ( md->params[i].type ) {
      case ChainPlus: cout << "c";
      case PlusEx:
      case Plus: cout << "+" << pd->params[i] << "." << node->params[i]; break;
      case MinEx: cout << "-" << pd->params[i] << "." << node->params[i]; break;
      case ChainConstant: cout << "c";
      case Constant:
        if ( showconstantnames )
          cout << "#" << typetable.getConstantName ( pd->params[i], node->params[i] );
        else
          cout << "#" << pd->params[i] << "." << node->params[i];
        break;
    }

    if ( i < md->size - 1 )
      cout << ", ";
  }
  cout << ")";

}

void printNode2 ( Node *node ) {
  ModeData *md = modetable.modeData ( MODE ( node ) );
  PredicateData *pd = predicatetable.predicateData ( md->predicate );
  cout << pd->name << "(";
  for ( int i = 0; i < md->size; i++ ) {
    switch ( md->params[i].type ) {
      case ChainPlus:
      case PlusEx:
      case Plus: cout << "V" << pd->params[i] << "N" << node->params[i]; break;
      case MinEx: cout << "V" << pd->params[i] << "N" << node->params[i]; break;
      case ChainConstant:
      case Constant:
          cout << typetable.getConstantName ( pd->params[i], node->params[i] );
        break;
    }

    if ( i < md->size - 1 )
      cout << ",";
  }
  cout << ")";

}

int nodeCmp ( Node *node1, Node *node2 ) {
  if ( node1->mode < node2->mode )
    return -1;
  else
    if ( node1->mode > node2->mode )
      return +1;
    else {
      INTTYPE *params1 = node1->params, *params2 = node2->params;
      int *inputvarpos = modetable.inputVarPos ( node1->mode );
      int nrinput = modetable.nrInput ( node1->mode ), i = 0, b = 0;
      while ( i < nrinput &&
              ( b = node1->params[inputvarpos[i]] - node2->params[inputvarpos[i]] ) == 0 )
	i++;
      return b;
    }
}


int nodeCmp2b ( TreeNodePointer *tnp1, TreeNodePointer *tnp2 ) {
  bool b1 = modetable.haschainpred ( tnp1->node->mode );
  bool b2 = modetable.haschainpred ( tnp2->node->mode );
  if ( b1 ) {
    if ( !b2 )
      return -1;
  }
  else
    if ( b2 )
      return +1;

  int diff = modetable.nrInput ( tnp1->node->mode ) - modetable.nrInput ( tnp2->node->mode );
  if ( diff > 0 )
    return +1;
  else
    if ( diff < 0 )
      return -1;

  int sc1 = tnp1->node->backtracksize * tnp2->node->support;
  int sc2 = tnp2->node->backtracksize * tnp1->node->support;

  // The code in the comments of this function provide some ideas for other orders.

  /*int sc1 = tnp1->node->backtracksize * tnp2->node->support * tnp2->node->support +
            tnp1->node->support * tnp2->node->support * tnp2->node->support;
  int sc2 = tnp2->node->backtracksize * tnp1->node->support * tnp1->node->support +
            tnp2->node->support * tnp1->node->support * tnp1->node->support;*/
  if ( sc1 > sc2 )
    return -1;
  if ( sc2 > sc1 )
    return +1;
  /*
  double ll1 = modetable.outputListLength ( tnp1->node->mode );
  double ll2 = modetable.outputListLength ( tnp2->node->mode );
  if ( ll1 > ll2 )
    return -1;
  if ( ll1 < ll2 )
    return +1;
    */
  /*int nrout1 = modetable.nrOutput ( tnp1->node->mode );
  int nrout2 = modetable.nrOutput ( tnp2->node->mode );
  if ( nrout1 > nrout2 )
    return -1;
  if ( nrout1 < nrout2 )
    return +1;*/
  if ( tnp1->node->support < tnp2->node->support )
    return -1;
  if ( tnp1->node->support > tnp2->node->support )
    return +1;
  if ( tnp1->hashnr < tnp2->hashnr )
    return -1;
  if ( tnp1->hashnr > tnp2->hashnr )
    return +1;
  return tnp1->pos - tnp2->pos;
}

int nodeCmp2 ( Node *node1, Node *node2 ) {
  bool b1 = modetable.haschainpred ( node1->mode );
  bool b2 = modetable.haschainpred ( node2->mode );
  if ( b1 ) {
    if ( !b2 )
      return -1;
  }
  else
    if ( b2 )
      return +1;

  int diff = modetable.nrInput ( node1->mode ) - modetable.nrInput ( node2->mode );
  if ( diff > 0 )
    return +1;
  else
    if ( diff < 0 )
      return -1;

/*  int sc1 = node1->backtracksize * node2->support * node2->support +
            node1->support * node2->support * node2->support;
  int sc2 = node2->backtracksize * node1->support * node1->support +
            node2->support * node1->support * node1->support;*/
  int sc1 = node1->backtracksize * node2->support;
  int sc2 = node2->backtracksize * node1->support;
  if ( sc1 > sc2 )
    return -1;
  if ( sc2 > sc1 )
    return +1;
  /*double ll1 = modetable.outputListLength ( node1->mode );
  double ll2 = modetable.outputListLength ( node2->mode );
  if ( ll1 > ll2 )
    return -1;
  if ( ll1 < ll2 )
    return +1;
  */
  /*int nrout1 = modetable.nrOutput ( node1->mode );
  int nrout2 = modetable.nrOutput ( node2->mode );
  if ( nrout1 > nrout2 )
    return -1;
  if ( nrout1 < nrout2 )
    return +1;*/
  if ( node1->support < node2->support )
    return -1;
  if ( node1->support > node2->support )
    return +1;
  return nodeCmp ( node1, node2 );
}

int nodeCmp3 ( const void *tnp1, const void *tnp2 ) {
  return nodeCmp2b ( (TreeNodePointer*) tnp1, (TreeNodePointer*) tnp2 );
}

TreeNodePointer ChildList::searchresult;

ChildList::ChildList () {
  modefirstchild = -1;
}

ChildList::~ChildList () {
}

TreeNodePointer *ChildList::search ( Node *node ) {
  if ( modefirstchild == -1 )
    return NULL;
  searchresult.hashnr = modetable.deepmode ( node->mode ) - modefirstchild;
  if ( searchresult.hashnr < 0 || searchresult.hashnr >= hash->size () )
    return NULL;
  HashNodes *hn = (*hash) [ searchresult.hashnr ];
  if ( !hn )
    return NULL;
  FArray<Node> &nodes = hn->nodes;
  int l = 0, r = nodes.size () - 1, re;
  
  while ( l <= r ) {
    searchresult.pos = ( l + r ) / 2;
    re = nodeCmp ( node, & ( nodes[searchresult.pos] ) );
    if ( re < 0 )
      r = searchresult.pos - 1;
    else
      if ( re > 0 )
        l = searchresult.pos + 1;
      else {
        searchresult.node = & ( nodes[searchresult.pos] );
        return &searchresult;
      }
  }
  return NULL;
}

TreeNodePointer *ChildList::searchbeforeequal ( Node *node, TreeNodePointer *before ) {
  // before must be a treenode pointer to a node in the current childlist.
  // we don't check this for reasons of efficiency, but this must be the case!
  int re = nodeCmp ( node, before->node );
  if ( re > 0 )
    return NULL;
  else
    if ( re == 0 )
      return before;
    else {
      searchresult.hashnr = modetable.deepmode ( node->mode ) - modefirstchild;
      if ( searchresult.hashnr < 0 || searchresult.hashnr >= hash->size () )
        return NULL;
  
      HashNodes *hn = (*hash) [ searchresult.hashnr ];
      if ( !hn )
	return NULL;
      FArray<Node> &nodes = hn->nodes;
      int l = 0, r, re;
      if ( searchresult.hashnr < modetable.deepmode ( before->node->mode ) - modefirstchild )
        r = nodes.size () - 1;
      else
        r = before->pos - 1;

      while ( l <= r ) {
	searchresult.pos = ( l + r ) / 2;
	re = nodeCmp ( node, & ( nodes[searchresult.pos] ) );
	if ( re < 0 )
	  r = searchresult.pos - 1;
	else
	  if ( re > 0 )
            l = searchresult.pos + 1;
	  else {
            searchresult.node = & ( nodes[searchresult.pos] );
            return &searchresult;
	  }
      }
      return NULL;
    }
}

FArray<Node> &ChildList::allocate ( int modenr, int size ) {
  if ( modefirstchild == -1 ) {
    modefirstchild = modenr;
    hash = new FArray<HashNodes*> ( ( modetable.size () - modenr ) / 2 + 1 );
      // scales automatically anyway, but this seems a good first choice.
  }
  HashNodes *hn = new HashNodes ( size );
  hash->insertAt ( modenr - modefirstchild, hn );
  
  return hn->nodes;
}

FArray<Node> *ChildList::nodess ( int deepmodenr ) {
  if ( modefirstchild == -1 ) 
    return NULL;
  HashNodes *hn = hash->element ( deepmodenr - modefirstchild );
  return hn ? &hn->nodes : NULL;
}

class ChildListIterator {
    TreeNodePointer tnp;
    int inhashmax;
    void gotonexthash () {
      do
        tnp.hashnr++;
      while ( tnp.hashnr < tnp.childlist->hash->size () && 
              (*tnp.childlist->hash) [tnp.hashnr] == NULL );
      if ( tnp.hashnr < tnp.childlist->hash->size () ) {
        tnp.pos = 0;
	inhashmax = (*tnp.childlist->hash) [tnp.hashnr]->nodes.size ();
      }
      else
        tnp.hashnr = -1;
    }
  public:
    ChildListIterator ( ChildList *childlist ) { 
      if ( childlist ) {
        tnp.childlist = childlist; 
        tnp.hashnr = -1; 
        gotonexthash (); 
      }
      else
        tnp.hashnr = -1;
    }
    TreeNodePointer *current () { 
      if ( tnp.hashnr == -1 ) 
        return NULL;
      else {
        tnp.node = & ( (*tnp.childlist->hash) [tnp.hashnr]->nodes[tnp.pos] );
        return &tnp;
      }
    }
    void next () {
      tnp.pos++;
      if ( tnp.pos >= inhashmax ) 
        gotonexthash ();
    }
};

TreeNodePointer roottnp;

NodeHistory::NodeHistory () { top = -1; }

void NodeHistory::init () {
  constappears = new bool*[typetable.size ()];
  variables = new int[typetable.size ()];
  variableoutputatom = new FArray<int>*[typetable.size ()];
  for ( int i = 0; i < typetable.size (); i++ ) {
    variables[i] = 0;
    constappears[i] = new bool[typetable.typeData ( i )->constnumbers];
    bzero ( constappears[i], sizeof(bool) * typetable.typeData ( i )->constnumbers );
    variableoutputatom[i] = new FArray<int>;
  }
  top = -1;
  nroutputatoms = 0;
  push ( roottnp );
}

void NodeHistory::push ( TreeNodePointer &tnp ) {
  ModeData *md = modetable.modeData ( tnp.node->mode );
  PredicateData *pd = predicatetable.predicateData ( md->predicate );
  pd->number--;
  top++;
  bool outputs = false;

  if ( !constappearshistory.element ( top ) )
    constappearshistory.insertAt ( top, new FArray<bool> ( md->size ) );
  for ( int i = 0; i < md->size; i++ ) {
    switch ( md->params[i].type ) {
      case MinEx:
        variables[md->params[i].number]++;
	variableoutputatom[md->params[i].number]->insertAt ( tnp.node->params[i], nroutputatoms );
	outputs = true;
	break;
      case Constant: case ChainConstant:
        constappearshistory [ top ] -> insertAt
	  ( i,
	    constappears[pd->params[i]][tnp.node->params[i]]
	  );
	constappears[pd->params[i]][tnp.node->params[i]] = true;
    }
  }

  if ( outputs ) {
    outputatoms.insertAt ( nroutputatoms, top );
    nroutputatoms++;
  }

  if ( nroutputatoms > 2 ) // 1: the key, 2: another atom, with 3 buys(X,Y1),buys(X,Y2) situations are possible
    equivalentpossible.insertAt ( top, true ); // better determinations possible.
  else
    if ( md->equivalentmodes.size () )
      equivalentpossible.insertAt ( top, true );
    else
      if ( top > 0 )
        equivalentpossible.insertAt ( top, equivalentpossible[top-1] );
      else
        equivalentpossible.insertAt ( 0, false );

  elements.insertAt ( top, tnp );
}

TreeNodePointer &NodeHistory::pop () {
  TreeNodePointer &p = elements[top];
  ModeData *md = modetable.modeData ( p.node->mode );
  PredicateData *pd = predicatetable.predicateData ( md->predicate );
  bool outputs = false;

  for ( int i = 0; i < md->size; i++ ) {
    switch ( md->params[i].type ) {
      case MinEx: variables[md->params[i].number]--; outputs = true; break;
      case Constant: case ChainConstant:
        constappears[pd->params[i]][p.node->params[i]] =
	  (* constappearshistory[top] ) [ i ];
    }
  }

  if ( outputs )
    nroutputatoms--;

  pd->number++;
  top--;
  return p;
}

void NodeHistory::print () {
  cout << setw ( 5 ) << last().node->support << " " << size () << " ";
  for ( int i = 0; i < size () - 1; i++ ) {
    printNode2 ( element ( i ).node );
    cout << ",";
  }
  printNode2 ( element ( size () - 1 ).node );
  cout << endl;
}

int counter = 0;

void NodeHistory::print2 () {
  cout << "mol(q" << counter << ",V0N0) :- ";
  //printNode2 ( element ( 0 ).node );
  //cout << " :- ";
  for ( int i = 1; i < size (); i++ ) {
    printNode2 ( element ( i ).node );
    if ( i < size () - 1 )
      cout << ",";
  }
  for ( int i = 0; i < typetable.size (); i++ ) {
    if ( variables[i] > 1 ) {
	for ( int k = 0; k < variables[i]; k++ ) {
	  for ( int t = k + 1; t < variables[i]; t++ )
	    cout << ",V" << i << "N" << k << "\\=" << "V" << i << "N" << t ;
	}
    }
  }
  cout << "." << endl;
  counter++;
}

NodeHistory nodeHistory;

// Contains data for the generation of dependent modes for the last
// node in the NodeHistory.
// This class stores data that can be used by all DependentDeepModeGenerators.
class NodeDependentGenerateData {
  public:
    Node *node;
    ModeData *md;
    int mdsize;
    int depth;
    int *dependentmodes;
    int depmodsize;
    ModeParamData *mpd;
    static int *varsinnode;
    static void staticInit () {
      varsinnode = new int[typetable.size ()];
    }
    static void staticCleanup () {
      delete[] varsinnode;
    }
   
    NodeDependentGenerateData () {
      int i;
      node = nodeHistory.last ().node;
      md = modetable.modeData ( MODE ( node ) );
      dependentmodes = md->dependentmodes;
      depmodsize = md->depmodsize;
      if ( !depmodsize ) 
        return;
      mpd = md->params;
      mdsize = md->size;
      // collect variables introduced by this node
      for ( i = 0; i < typetable.size (); i++ )
        varsinnode[i] = 0;
      for ( i = 0; i < mdsize; i++, mpd++ )
        if ( mpd->type == MinEx )
          varsinnode[mpd->number]++;
    }
};

int *NodeDependentGenerateData::varsinnode;


// a generator for dependent modes of the given mode type, given the
// query data stored in ndgd.
class DependentModeGenerator {
  private:
    static Node newnode; // the node to be generated.
    NodeDependentGenerateData &ndgd;
    static INTTYPE *vars, *nrofvar, *mustnrvar; // *outputvar;
    int *outputvarpos;
    int lastmustpar, firstmustpar;
    bool usesmust;
    int varparnr, j;
    bool stop;
  public:
    ModeData *newmd;
    static void staticInit () {
      //outputvar = new int[typetable.size ()];
      vars = new INTTYPE[predicatetable.maxNParams ()];
      nrofvar = new INTTYPE[predicatetable.maxNParams ()];
      mustnrvar = new INTTYPE[predicatetable.maxNParams ()];
      newnode.params = vars;
      newnode.freqchildren = NULL;
      //newnode.flags = FLAGSNONE;
      newnode.support = 0;
    }
    static void staticCleanup () {
      //delete[] outputvar;
      delete[] vars;
      delete[] nrofvar;
      delete[] mustnrvar;
    }
    DependentModeGenerator ( NodeDependentGenerateData &ndgd, ModeData *newmd ):
      ndgd ( ndgd ), newmd ( newmd ) {
      ModeParamData *mpd = newmd->params;
      outputvarpos = newmd->inputvarpos;
      int parsize = newmd->size, number, i;
      int *variables = nodeHistory.queryVariableCounts ();
      varparnr = 0;
      
      //newnode.introdepth = nodeHistory.size ();
      newnode.mode = newmd->nr;
      
      //for ( i = 0; i < typetable.size (); i++ )
      //  outputvar[i] = 0;

      // possibilities for each parameter: constant, variable?
      for ( i = 0, mpd = newmd->params; i < parsize; i++, mpd++ ) {
        number = mpd->number;
        switch ( mpd->type ) {
          case Constant: {
            vars[i] = number;
            break;
          }
          case ChainConstant: {
            vars[i] = nodeHistory.back ( mpd->back + 1 ).node->params[mpd->backparamnr];
            break;
          }
          case MinEx: {
            //vars[i] = variables[number] + ndgd.varsinnode[number] + ( outputvar[number]++ );
	    //fill in can be generalized later.
            break;
          }
          case ChainPlus: {
            vars[i] = nodeHistory.back ( mpd->back + 1 ).node->params[mpd->backparamnr];
            break;
          }
          case Plus: case PlusEx: {
            int number = newmd->params[i].number;
            nrofvar[varparnr] = variables[number] - 1;
            mustnrvar[varparnr] = ndgd.varsinnode[number];
	    if ( mustnrvar[varparnr] > 0 )
	      lastmustpar = varparnr;
            varparnr++;
          }
        }
      }
    }
    int number () const {
      int i, j, prod, nodessize;
      if ( !newmd->haschainpred ) {
        nodessize = 0;
      	for ( i = 0; i < varparnr; i++ ) {
          prod = 1;
          for ( j = 0; j < i; j++ )
	    prod *= nrofvar[j] + 1 - mustnrvar[j];
          prod *= mustnrvar[j];
          for ( j++; j < varparnr; j++ )
            prod *= nrofvar[j] + 1;
          nodessize += prod;
	}
        return nodessize;
      }
      else {
        if ( varparnr == 0 )
	  return 1;
        nodessize = 1;
        for ( j = 0; j < varparnr; j++ )
          nodessize *= nrofvar[j] + 1;
	return nodessize;
      }
    }
    void start () {
      stop = false;
      usesmust = false;
      if ( !newmd->haschainpred ) {
        if ( lastmustpar == 0 ) {
	  usesmust = true;
	  vars[outputvarpos[0]] = nrofvar[0] - mustnrvar[0] + 1;
	}
	else
          vars[outputvarpos[0]] = 0;
        j = 0;
	gotonext_nochain ();
      }
      else {
        if ( varparnr == 0 )
	  return;
	vars[outputvarpos[0]] = 0;
        j = 0;
	gotonext_chain ();
      }
    }
    void next () {
      if ( !newmd->haschainpred ) {
        if ( varparnr == 0 ) {
	  stop = true;
	  return;
	}
        j--;
	if ( j < 0 )
	  stop = true;
	else {
          vars[outputvarpos[j]]++;
	  gotonext_nochain ();
	}
      }
      else {
        if ( varparnr == 0 ) {
	  stop = true;
	  return;
	}
        if ( j < 1 )
	  stop = true;
	else {
	  if ( j == firstmustpar ) 
	    usesmust = false;
          j--;
          vars[outputvarpos[j]]++;
	  if ( !usesmust && vars[outputvarpos[j]] >= nrofvar[j] - mustnrvar[j] + 1 ) {
	    usesmust = true;
	    firstmustpar = j;
	  }
          gotonext_chain ();
	}
      }
    }
    inline Node *current () const { if ( stop ) return NULL; else return &newnode; }
  private:
    void gotonext_nochain () {
      while ( true ) {
        if ( vars[outputvarpos[j]] <= nrofvar[j] ) {
          j++; // goto next variable
          if ( j == varparnr )
	    return;
          else
	    if ( !usesmust && lastmustpar == j )
              vars[outputvarpos[j]] = nrofvar[j] - mustnrvar[j] + 1;
	    else {
	      vars[outputvarpos[j]] = 0;
	      if ( !usesmust && nrofvar[j] - mustnrvar[j] + 1 == 0 ) {
		usesmust = true;
		firstmustpar = j;
	      }
	    }
        }
        else {
	  if ( j == firstmustpar ) 
	    usesmust = false;
          j--;
	  if ( j < 0 )
	    break;
          vars[outputvarpos[j]]++;
	  if ( !usesmust && vars[outputvarpos[j]] >= nrofvar[j] - mustnrvar[j] + 1 ) {
	    usesmust = true;
	    firstmustpar = j;
	  }
        }
      }
      stop = true;
    }
    void gotonext_chain () {
      while ( true )
        if ( vars[outputvarpos[j]] <= nrofvar[j] ) {
          j++; // goto next variable
          if ( j == varparnr )
	    return;
          else
            vars[outputvarpos[j]] = 0;
        }
        else {
          j--;
	  if ( j < 0 )
	    break;
          vars[outputvarpos[j]]++;
        }
      stop = true;
    }
};

Node DependentModeGenerator::newnode; // the node to be generated.
INTTYPE *DependentModeGenerator::vars, *DependentModeGenerator::nrofvar, *DependentModeGenerator::mustnrvar; // *outputvar;

class DependentDeepModeGenerator {
  private:
    DependentModeGenerator *dmg;
    NodeDependentGenerateData &ndgd;
    int i, i2, j, size, _number, depsize, firstmodenr;
  public:
    DependentDeepModeGenerator ( NodeDependentGenerateData &ndgd, int firstmodenr ): 
      ndgd ( ndgd ), firstmodenr ( firstmodenr ), dmg ( NULL ) {
    }
    ~DependentDeepModeGenerator () {
      if ( dmg )
        delete dmg;
    }
    void start () {
      i = 0;
      j = 0;
      ModeData *md = modetable.modeData ( firstmodenr );
      PredicateData *pd = predicatetable.predicateData ( md->predicate );
      if ( pd->number > 0 ) {
        size = md->activefrequent.activeSize ();
	if ( size == 0 ) {
	  _number = 0;
	  return;
	}
	i2 = md->activefrequent.firstActive ();
	if ( i2 != 0 )
	  md = modetable.modeData ( firstmodenr + i2 );
	i2 += firstmodenr;
        dmg = new DependentModeGenerator ( ndgd, md );
        depsize = dmg->number ();
        _number = size * depsize;
        if ( depsize ) {
          dmg->start ();
	  checkplusex ();
	}
      }
      else
        _number = 0;
    }
    Node *current () {
      if ( i < size && _number )
        return dmg->current ();
      else
        return NULL;
    }
    void next () {
      if ( ++j < depsize ) {
        dmg->next ();
        checkplusex ();
      }
      else {
        delete dmg;
	i++;
	if ( i < size ) {
	  ModeData *md = modetable.nextFrequentActive ( i2 );
	  dmg = new DependentModeGenerator ( ndgd, md );
	  j = 0;
	  dmg->start ();
          checkplusex ();
	}
	else
	  dmg = NULL;
      }
    }    
    int number () const { return _number; }
    bool violatesplusex ( Node *node ) {
      for ( int i = 0; i < dmg->newmd->size; i++ ) {
        if ( dmg->newmd->params[i].type == PlusEx ) {
	  for ( int j = 0; j < dmg->newmd->size; j++ )
	    if ( i != j && dmg->newmd->params[j].type == PlusEx &&
		 node->params[i] == node->params[j] )
	      return true;
	}
      }
      return false;
    }
    void checkplusex () {
      if ( dmg && dmg->newmd->nrplusex ) {
        Node *node;
        while ( ( node = current () ) && violatesplusex ( node ) )
	  next ();
      }
    }
};

struct NewNodeData {
  Node *node;  // the dependent node
  int bucket;  // the bucket into which this node belongs.
  int bucketsize; // the size of the bucket, as estimated by the generator.
};

class DependentGenerator {
  private:
    NodeDependentGenerateData ndgd;
    int i;
    DependentDeepModeGenerator *ddmg;
    static NewNodeData nnd;
    void gotonext () {
      while ( i < ndgd.depmodsize && !( nnd.node = ddmg->current () ) ) {
        delete ddmg;
	i++;
	if ( i < ndgd.depmodsize ) {
	  ddmg = new DependentDeepModeGenerator ( ndgd, ndgd.dependentmodes[i] );
	  ddmg->start ();
	  nnd.bucket = modetable.deepmode ( ndgd.dependentmodes[i] );
	  nnd.bucketsize = ddmg->number ();
	}
	else
	  ddmg = NULL;
      }
    }
  public:
    DependentGenerator () { ddmg = NULL; }
    ~DependentGenerator () { if ( ddmg ) delete ddmg; }
    void start () {
      i = 0;
      if ( ndgd.depmodsize ) {
        ddmg = new DependentDeepModeGenerator ( ndgd, *ndgd.dependentmodes );
	ddmg->start ();
	nnd.bucket = modetable.deepmode ( ndgd.dependentmodes[i] );
        nnd.bucketsize = ddmg->number ();
	gotonext ();
      }
    }
    void next () {
      ddmg->next ();
      gotonext ();
    }
    inline NewNodeData *current () const {
      if ( i < ndgd.depmodsize ) 
        return &nnd;
      else 
        return NULL;
    }
};

NewNodeData DependentGenerator::nnd;

class BrotherGenerator {
  private:
    static NewNodeData nnd;
    ChildList *childlist;
    int hashnr, inhashpos;
    int chainprednr;
    ModeData *md;
    Node *sourcenode;

    void gotonextchild () {
      do
        hashnr++;
      while ( hashnr < childlist->hash->size () && 
              (*childlist->hash) [hashnr] == NULL );
      if ( hashnr < childlist->hash->size () ) {
        inhashpos = 0;
	nnd.bucketsize = (*childlist->hash) [hashnr]->nodes.size ();
	nnd.bucket = hashnr + childlist->modefirstchild;
      }
      else
        hashnr = -1;
    }
    inline bool validBrother () const {
      Node &node = (*childlist->hash) [hashnr]->nodes[inhashpos];
      ModeData *md2 = modetable.modeData ( node.mode );
      PredicateData *pd2 = predicatetable.predicateData ( md2->predicate );
      int cmp = nodeCmp2 ( &node, sourcenode );
      if ( ( cmp > 0 || ( cmp == 0 && md2->nroutput && !md2->primarykeys ) ) 
           && ( chainprednr == -1 || !md2->haschainpred || md2->chainprednr != chainprednr )
           && pd2->number > 0 && !pd2->key ) {
	nnd.node = &node;
        return true;
      }
      else {
        return false;
      }
    }
    void gotovalid () {
      while ( hashnr != -1 && !validBrother () ) {
	inhashpos++;
        if ( inhashpos == nnd.bucketsize )
	  gotonextchild ();
      }
    }
  public:
    BrotherGenerator () { 
      sourcenode = nodeHistory.last().node;
      childlist = nodeHistory.last ().childlist; 
      hashnr = -1;
      md = modetable.modeData ( sourcenode->mode );
      chainprednr = md->haschainpred ? md->chainprednr : -1;
      gotonextchild ();
      gotovalid ();
      if ( hashnr == -1 )
        return;
      
      HashNodes *hn = childlist->hash->element ( hashnr );
      if ( hn == NULL ) {
        hashnr = -1;
	return;
      }
      nnd.bucketsize = hn->nodes.size ();
      nnd.bucket = hashnr + childlist->modefirstchild;
    }
    ~BrotherGenerator () { }
    void start () {
      // does nothing. Just to be consistent...
    }
    void next () { 
      inhashpos++;
      if ( inhashpos == nnd.bucketsize )
        gotonextchild ();
      gotovalid ();
    }
    NewNodeData *current () {
      if ( hashnr == -1 )
        return NULL;
      else
        return &nnd;
    }
};
    
QueryExpansions::QueryExpansions () {
}

void QueryExpansions::initialize () {
  activelist = new ActiveList[modetable.deepsize ()];
  brothertidwalk = new FArray < FListWalk < TidData > > [ modetable.deepsize () ];
  for ( int i = 0; i < modetable.deepsize (); i++ )
    candidates.allocate ( i, 10 /* some reasonable number */ );
}

QueryExpansions::~QueryExpansions () {
  delete[] activelist;
  delete[] brothertidwalk;
}

void QueryExpansions::init () {
  for ( int i = 0; i < modetable.deepsize (); i++ ) {
    (*candidates.hash) [i]->nodes.resize ( 0 );
    brothertidwalk[i].resize ( 0 );
  }
  if ( modetable.alwaystrue ( nodeHistory.last().node->mode ) ) {
    alwaystrue = true;
    initactivesize = modetable.cdepsize;
  }
  else {
    alwaystrue = false;
    initactivesize = modetable.deepsize ();
  }
}

// the parameter node should have tidlist filled in to a tid list that should be taken into account
// while counting children.
void QueryExpansions::addNode ( ModeData *md, Node *newnode ) {
  FArray<Node> &nodes = (*candidates.hash) [md->deepnr]->nodes;
  brothertidwalk[md->deepnr].resize ( nodes.size () + 1 );
  if ( alwaystrue )  {
    if ( newnode->tidlist )
      newnode->tidlist = newnode->tidlist->copy ();
    brothertidwalk[md->deepnr][nodes.size ()].setTo ( NULL ); // just to make sure
  }
  else {
    if ( newnode->tidlist ) {
      brothertidwalk[md->deepnr][nodes.size ()].setTo ( &newnode->tidlist->tids );
    }
    else
      brothertidwalk[md->deepnr][nodes.size ()].setTo ( NULL );
    newnode->tidlist = NULL;
  }
  nodes.append ( *newnode );
}

void QueryExpansions::setTid ( int tid ) {
  activeactivelist.setActive ( initactivesize );
  for ( int i = 0; i < initactivesize; i++ ) {
    FArray<Node> &nodes = (*candidates.hash) [i]->nodes;
    int size = nodes.size ();
    activelist[i].setActive ( size );
    for ( int j = 0; j < size; j++ ) {
      // inactivate candidates that were shown in a brother not to be contained
      // in this query.
      if ( modetable.quickpreeval ( nodes[j].mode ) ) {
        // we can determine quickly whether this constants and this predicate
	// occur in this example, and otherwise discard this candidate
	// immediately, preventing backtracking in this way.
	if ( !modetable.computeQuickPreEval ( nodes[j].mode, tid, nodes[j].params ) ) {
	  activelist[i].setInActive ( j );
	  continue; // skip the remainder of this for loop.
	}
      }
      if ( brothertidwalk[i][j].current () ) {
        while ( brothertidwalk[i][j].current () && brothertidwalk[i][j].current ()->tid < tid )
	  brothertidwalk[i][j].next ();
	if ( !brothertidwalk[i][j].current () || brothertidwalk[i][j].current ()->tid != tid )
	  activelist[i].setInActive ( j );
      }
    }
    if ( activelist[i].activeSize () == 0 )
      activeactivelist.setInActive ( i );
  }
}

void QueryExpansions::firstActive () {
  nextbucket = currentbucket = activeactivelist.firstActive ();
  if ( currentbucket != -1 ) {
    currentpos = activelist[currentbucket].firstActive ();
    if ( currentpos != -1 ) {
      nextpos = activelist[currentbucket].nextActive ( currentpos );
      if ( nextpos == -1 ) {
	nextbucket = activeactivelist.nextActive ( currentbucket );
	if ( nextbucket != -1 )
	  nextpos = activelist[nextbucket].firstActive ();
      }
    }
  }
}

Node *QueryExpansions::currentActive () {
  if ( currentbucket == -1 )
    return NULL;
  return & (*candidates.hash) [currentbucket]->nodes[currentpos];
}

void QueryExpansions::setCurrentInActive () {
  activelist[currentbucket].setInActive ( currentpos);
  if ( activelist[currentbucket].activeSize () == 0 )
    activeactivelist.setInActive ( currentbucket );
}

void QueryExpansions::nextActive () {
  currentpos = nextpos;
  currentbucket = nextbucket;
  if ( currentbucket != -1 ) {
    // we may have stored the end-of-walking marker in the nextpos (now currentpos)
    nextpos = activelist[currentbucket].nextActive ( currentpos );
    if ( nextpos == -1 ) {
      nextbucket = activeactivelist.nextActive ( currentbucket );
      if ( nextbucket != -1 )
	nextpos = activelist[nextbucket].firstActive ();
    }
  }
}

NewNodeData BrotherGenerator::nnd;

QueryExpansions queryexpansions;
QueryEvaluator queryevaluator;

int totalcount = 0, equalcount = 0, infrequnioncount = 0, infreqsubcount = 0, equivalentcount = 0;

class Generator {
  private:
    static int *newvar;
    static Node newnode;
  public:
    static void staticInit () {
      newvar = new int[typetable.size ()];
    }
    static void staticCleanup () {
      delete[] newvar;
    }
    
    int childcount;
    
    void makeNode ( ModeData *md, Node *node ) {
      newnode.mode = node->mode;
      newnode.params = new INTTYPE[md->size];
      newnode.freqchildren = newnode.infreqchildren = NULL;
      newnode.maxdepth = nodeHistory.size () + 1;
      int *variables;
      if ( md->nroutput ) {
	variables = nodeHistory.queryVariableCounts ();
	bzero ( newvar, sizeof(int) * typetable.size () );
      }
      for ( int i = 0; i < md->size; i++ ) {
	if ( md->params[i].type == MinEx ) {
	  int type = md->params[i].number;
	  newnode.params[i] = variables[type] + newvar[type];
	  newvar[type]++;
	}
	else
	  newnode.params[i] = node->params[i];
      }
      
      totalcount++;
    }
    
    void append_brother ( Node *node ) {
      ModeData *md = modetable.modeData ( node->mode );
      makeNode ( md, node );
      
      if ( searchforequalnode ( &newnode ) ) {
        equalcount++;
        delete[] newnode.params;
	return;
      }
      
      if ( frequentTidUnion ( nodeHistory.last().node->tidlist, node->tidlist ) ) {

        TreeNodePointer tnp;
	tnp.node = &newnode;
	tnp.childlist = &queryexpansions.candidates;
	nodeHistory.push ( tnp );
	
	// check first, or after computing join????
	
	if ( !searchforequivalentquery () || keepequivalent ) {
          newnode.tidlist = node->tidlist; // the single difference with append_dependent!!!
          if ( keepequivalent ) 
	    newnode.support = keepequivalent - 1;
	  else
	    newnode.support = childcount;
	  childcount++;
	  queryexpansions.addNode ( modetable.modeData ( node->mode ), &newnode );
          nodeHistory.pop ();
 	}
	else {
	  // Be aware! in this way, some equivalent nodes which have already
	  // been pruned by frequentTidUnion are marked infrequent but not
	  // completely removed!
          nodeHistory.pop ();
	  delete[] newnode.params;
        }
      }
      else
        infrequnioncount++;
    }
    
    void append_dependent ( Node *node ) {
      ModeData *md = modetable.modeData ( node->mode );
      makeNode ( md, node );
      
      if ( searchforequalnode ( &newnode ) ) {
        equalcount++;
        delete[] newnode.params;
	newnode.tidlist = NULL;
	return;
      }

      TreeNodePointer tnp;
      tnp.node = &newnode;
      tnp.childlist = &queryexpansions.candidates;

      nodeHistory.push ( tnp );
      
      if ( !searchforequivalentquery () || keepequivalent ) {
        newnode.tidlist = NULL;
        if ( keepequivalent ) 
          newnode.support = keepequivalent - 1;
	else
	  newnode.support = childcount;
	childcount++;
	queryexpansions.addNode ( modetable.modeData ( node->mode ), &newnode );
        nodeHistory.pop ();
      }
      else {
        nodeHistory.pop ();
        delete[] newnode.params;
      }
      
    }


    void copy ( DependentGenerator &dg ) {
      NewNodeData *nnd = dg.current ();
      int bucket = nnd->bucket;

      while ( nnd && nnd->bucket == bucket ) {
        append_dependent ( nnd->node );
	dg.next ();
	nnd = dg.current ();
      }
    }

    void copy ( BrotherGenerator &bg ) {
      NewNodeData *nnd = bg.current ();
      int bucket = nnd->bucket;
      
      while ( nnd && nnd->bucket == bucket ) {
	append_brother ( nnd->node );
	bg.next ();
	nnd = bg.current ();
      }
    }
    
    void merge ( DependentGenerator &dg, BrotherGenerator &bg ) {
      NewNodeData *nnd1 = bg.current (), *nnd2 = dg.current ();
      int bucket = nnd1->bucket, bucketsize = nnd1->bucketsize + nnd2->bucketsize;
      int cmp;
      
      while ( nnd1 && nnd2 && nnd1->bucket == bucket && nnd2->bucket == bucket ) {
        cmp = nodeCmp ( nnd1->node, nnd2->node );
	if ( cmp < 0 ) {
	  append_brother ( nnd1->node );
	  bg.next ();
	  nnd1 = bg.current ();
	}
	else {
	  //  == 0 should not occur.
	  append_dependent ( nnd2->node );
	  dg.next ();
	  nnd2 = dg.current ();
	}
      }
      while ( nnd1 && nnd1->bucket == bucket ) {
	append_brother ( nnd1->node );
	bg.next ();
	nnd1 = bg.current ();
      }
      while ( nnd2 && nnd2->bucket == bucket ) {
	append_dependent ( nnd2->node );
	dg.next ();
	nnd2 = dg.current ();
      }
    }

    void generate () {
      DependentGenerator dg;
      BrotherGenerator bg;
      NewNodeData *ndg, *nbg;
      queryexpansions.init ();
      childcount = 0;

      dg.start ();
      bg.start ();
      
      while ( ( ndg = dg.current () ) && ( nbg = bg.current () ) ) {
	if ( ndg->bucket < nbg->bucket )
	  copy ( dg );
	else
	  if ( ndg->bucket > nbg->bucket )
	    copy ( bg );
	  else
	    merge ( dg, bg );
      }
      while ( ndg = dg.current () ) 
        copy ( dg );
      while ( nbg = bg.current () )
        copy ( bg );
	
    }
    
    void count () {
      Node *node = nodeHistory.last().node;
      TidList *tidlist = node->tidlist;
      FListWalk<TidData> tidwalk ( &tidlist->tids );
      TidData *tidd;
      while ( tidd = tidwalk.current () ) {
        queryexpansions.setTid ( tidd->tid );
	queryevaluator.evaluate ( fulltidarray->tids[tidd->tid], &queryexpansions ); // maintains the tidlists of all candidates.
        tidwalk.next ();
      }
    }
    
    void copytonode () {
      // copy candidates as children of the last node.
      Node *expnode = nodeHistory.last().node;
      expnode->freqchildren = expnode->infreqchildren = NULL;

      FArray<TreeNodePointer> *order;
      if ( ! ( order = nodeHistory.nodeorders.element ( nodeHistory.size () ) ) ) {
        order = new FArray<TreeNodePointer>;
	nodeHistory.nodeorders.insertAt ( nodeHistory.size (), order );
      }
      else
        order->resize ( 0 );
      TreeNodePointer ordertnp;
      int childcount2 = 0;
      static FArray<int> sortmap;


      for ( int i = 0; i < modetable.deepsize (); i++ ) {
        FArray<Node>& nodes = (*queryexpansions.candidates.hash) [ i ]->nodes;
	if ( nodes.size () ) {
          FArray<Node> *freqchildren = NULL, *infreqchildren = NULL;
	  for ( int j = 0; j < nodes.size (); j++ ) {
	    Node &node = nodes[j];
	    if ( !node.tidlist || node.tidlist->tids.size () < minsup ) {
	      // add to infrequent children.
	      if ( nodeHistory.size () == 1 ) {
	        modetable.inactivateAllEquivalent ( &node );
	      }

	      if ( !expnode->infreqchildren )
	        expnode->infreqchildren = new ChildList;
	      if ( !infreqchildren )
	        infreqchildren = &expnode->infreqchildren->allocate ( modetable.deepmode ( node.mode ), nodes.size () );
	      if ( node.tidlist ) {
	        node.tidlist->kill ();
	      }
	      node.tidlist = NULL;
              infreqchildren->append ( node );
	    }
	    else {
	      // add to frequent children.
	      if ( !expnode->freqchildren )
	        expnode->freqchildren = new ChildList;
	      if ( !freqchildren )
	        freqchildren = &expnode->freqchildren->allocate ( modetable.deepmode ( node.mode ), nodes.size () );
  	      int orderpos; // support is temporarily filled
	                    // with a number that unique identifies an equivalency
			    // class in the set of children.
			    // equivalence class numbers are assigned as follows:
			    // - for every node which has an equivalent node in
			    //   childlist order, the number stored in that equivalent
			    //   node is stored here too.
			    // - otherwise, a new number is stored.
	      if ( node.support < childcount2 ) {
	        int oldsup = node.support;
	        node.support = node.tidlist->tids.size ();
		node.backtracksize = node.tidlist->backtracksize;
	        int cmp = nodeCmp2 ( &node, (*order) [ sortmap[oldsup] ].node );
		if ( cmp < 0 ) {
		  orderpos = oldsup;
		}
		else {
		  orderpos = -1;
		}
	      }
	      else {
	        orderpos = node.support;
		node.support = node.tidlist->tids.size ();
		node.backtracksize = node.tidlist->backtracksize;
	        sortmap.insertAt ( orderpos, order->size () );
              }

              freqchildren->append ( node );

	      if ( orderpos != -1 ) {
	        ordertnp.childlist = expnode->freqchildren;
	        ordertnp.hashnr = modetable.deepmode ( node.mode ) - expnode->freqchildren->modefirstchild;
	        ordertnp.pos = freqchildren->size () - 1;
	        ordertnp.node = & (*freqchildren) [ freqchildren->size () - 1 ];
	        order->insertAt ( sortmap[orderpos], ordertnp );
              }
	    }
	    childcount2++;
	  }
	  if ( freqchildren )
	    freqchildren->compact ();
	  if ( infreqchildren )
	    infreqchildren->compact ();
	}
      }

      // we have filled in the order array, which has to be sorted to determine
      // the order in which nodes are evaluated.
      qsort ( order->array, order->size (), sizeof ( TreeNodePointer ), nodeCmp3 );
    }

    void expand () {
      generate ();
      count ();
      copytonode ();
    }
    
    void init () {
      
    }
};

int *Generator::newvar;
Node Generator::newnode;


void initRootTNP () {
  roottnp.childlist = new ChildList;
  roottnp.hashnr = key->deepnr;
  roottnp.pos = 0;
  FArray<Node> &root = roottnp.childlist->allocate ( roottnp.hashnr, 1 );
  Node rootnode;
  rootnode.mode = key->nr;
  rootnode.params = new INTTYPE[key->size];
  rootnode.params[0] = 0;
  rootnode.tidlist = new TidList;
  rootnode.freqchildren = rootnode.infreqchildren = NULL;
  rootnode.maxdepth = 254;
  TypeData *td = typetable.typeData ( key->params[0].number );  
  rootnode.support = td->constnumbers;
  TidData tidd;
  
  for ( int i = 0; i < td->constnumbers; i++ ) {
    tidd.tid = i; // tidd is copied
    rootnode.tidlist->tids.append ( tidd );
  }

  root.append ( rootnode );
  roottnp.node = & ( root[0] );
}

void cleanupRootTNP () {
  delete[] roottnp.node->params;
  delete roottnp.childlist;
}


void staticInit () {
  initRootTNP ();
  NodeDependentGenerateData::staticInit ();
  DependentModeGenerator::staticInit ();
  Generator::staticInit ();
}

void staticCleanup () {
  NodeDependentGenerateData::staticCleanup ();
  DependentModeGenerator::staticCleanup ();
  Generator::staticCleanup ();
}

int freqcount = 0;
FArray<int> freqcounts;

int maxdepth = 0;

void print ( int depth ) {
  Node *node;
  node = nodeHistory.last().node;
  maxdepth = max ( depth, maxdepth );

  switch ( showstyle ) {
    case Queries: nodeHistory.print (); break;
    case Trie: for ( int i = 0; i < depth; i++ )
                 cout << "  ";
               printNode2 ( node );
	       cout << " (" << node->support << ")" << endl;
	       break;
    case ModeTrie: for ( int i = 0; i < depth; i++ )
                     cout << "  ";
                   printNode ( node );
		   cout << " (" << node->support << ")" << endl;
		   break;
  }

  Generator g;
  g.expand ();

  int nodemaxdepth = node->maxdepth;

  FArray<TreeNodePointer> *sortedlist = nodeHistory.nodeorders[nodeHistory.size()];
  TreeNodePointer *n;

  if ( !modetable.haschainsucc ( node->mode ) ) {
    if ( depth >= freqcounts.size () )
      freqcounts.insertAt ( depth, 0 );
    freqcounts[depth]++;
    freqcount++;


    for ( int i = 0; i < sortedlist->size (); i++ ) {
      n = & (*sortedlist) [i]; // additional test occured here (support>0).
       // this test should not be necessary as support==0 nodes are not
       // put into the sorted list.
      nodeHistory.push ( *n );
      applyAllAssignments ();
      print ( depth + 1 );
      undoAllAssignments ();
      nodeHistory.pop ();
#ifdef DIRTY
      if ( depth == 0 )
        modetable.inactivateAllEquivalent ( n->node );
#endif
      if ( n->node->maxdepth > nodemaxdepth )
        nodemaxdepth = n->node->maxdepth;
    }
  }
  else {
    int deepnr = modetable.deepmode ( node->mode );

    for ( int i = 0; i < sortedlist->size (); i++ ) {
      n = & (*sortedlist) [i];
      if ( modetable.chainprednr ( n->node->mode ) == deepnr ) {
        nodeHistory.push ( *n );
        applyAllAssignments ();
        print ( depth + 1 );
        undoAllAssignments ();
        nodeHistory.pop ();
#ifdef DIRTY
        if ( depth == 0 )
          modetable.inactivateAllEquivalent ( n->node );
#endif
        if ( n->node->maxdepth > nodemaxdepth )
          nodemaxdepth = n->node->maxdepth;

      }
    }
  }

  node->maxdepth = nodemaxdepth;
  if ( node->tidlist )
    node->tidlist->kill ();
}

FullTidArray *fulltidarray;

void print () {
  staticInit ();
  queryexpansions.initialize ();
  queryevaluator.initialize ();
  nodeHistory.init ();
  fulltidarray = new FullTidArray ();
  
  print ( 0 );
  
  delete fulltidarray;

  staticCleanup ();
  OUT << "[    Query   Number of    ]" << endl
      << "[    Length  Queries      ]" << endl
      << "[    ------  ----------   ]" << endl;
  for ( int k = 0; k < freqcounts.size (); k++ )
    OUT << "[    " << setw ( 6 ) << k + 1 << "  " << setw ( 10 ) << freqcounts[k] << "   ]" << endl;
  OUT << "[            ---------- + ]" << endl
      << "[            " << setw ( 10 ) << freqcount << "   ]" << endl;
  OUT_F
}
