/***************************************************************************
                          treeprune.cpp  - determines whether
			                   an equivalent query exists
					   earlier in the trie
                             -------------------
    begin                : Tue Jan 21 2003
    copyright            : (C) 2003 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "tree.h"
#include "globaldata.h" 
#include <iostream>

using namespace std;

bool sharepkey ( Node *node1, ModeData *mode1, Node *node2, ModeData *mode2, FArray<int> *pkey ) {
  for ( int i = 0; i < pkey->size (); i++ ) {
    int l = (*pkey) [ i ];
    if ( ( mode2->params[l].type == Constant ||
           mode2->params[l].type == ChainConstant ) ) {
      if ( ( mode1->params[l].type == Constant ||
	     mode1->params[l].type == ChainConstant ) ) {
	if ( node1->params[l] != node2->params[l] )
          return false;
      }
      else
        return false;
    }
    else
      if ( mode2->params[l].type != MinEx ) // a plus
        if ( mode1->params[l].type != Constant && 
	     mode1->params[l].type != ChainConstant ) {
           if ( node1->params[l] != node2->params[l] )
	     return false;
	}
	else // a constant
	  return false;
  }
  
  return true;
}

bool hasequalpkey ( Node *node1, ModeData *mode1, Node *node2, ModeData *mode2 ) {
  FListWalk<FArray<int> *> walk ( mode2->primarykeys );
  FArray<int> **current;
  
  while ( current = walk.current () ) {
    if ( sharepkey ( node1, mode1, node2, mode2, *current ) )
      return true;
    walk.next ();
  }
  
  return false;
}

// is node1 equal to node2? 
bool isequalnode ( Node *node1, ModeData *mode1, Node *node2, ModeData *mode2 ) {
  if ( mode1->predicate != mode2->predicate )
    return false;
  if ( mode2->primarykeys )
    if ( hasequalpkey ( node1, mode1, node2, mode2 ) )
      return true;
  if ( mode1->nrconst == mode2->nrconst ) {
    // now the choice is difficult. We could use the `equivalent' information
    // that we collected while reading the modes. We could also compare
    // the atoms themselves. 
    // For the equivalent information we have to walk through the list of equivalent
    // modes to search for the mode we are comparing with --> a for loop.
    // For direct comparison we have to walk through the parameters --> a for loop
    // Which of these two is shorter?
    // We assume for now that the difference is is not signigicant and
    // the number of parameters is small. 
    ModeParamData *mpd1 = mode1->params,
                  *mpd2 = mode2->params;
    INTTYPE *np1 = node1->params, *np2 = node2->params;
    int n = mode1->size;
    for ( int i = 0; i < n; i++, mpd1++, mpd2++, np1++, np2++ )
      if ( IS_VAR_PARAM ( mpd1 ) ) {
        if ( IS_VAR_PARAM ( mpd2 ) ) {
	  if ( *np1 != *np2 )
	    return false;
	}
	else
	  return false;
      }
      else
        if ( IS_VAR_PARAM ( mpd2 ) )
	  return false;
	else
	  if ( *np1 != *np2 )
	    return false;
    return true;
  }
  return false;
}

// searches for an atom in the history which is exactly the same.
// newnode should not be in the nodehistory.
bool searchforequalnode ( Node *newnode ) {
  ModeData *md = modetable.modeData ( newnode->mode );
  if ( md->equivalentmodes.size () == 0 ) {
    if ( md->primarykeys ) {
      Node *node2 = nodeHistory.element ( nodeHistory.size () - 1 ).node;
      if ( newnode->mode == node2->mode &&
           hasequalpkey ( node2, 
                          modetable.modeData ( MODE ( node2 ) ),
			  newnode,
			  md ) )
        return true;
    }
    return false;
  }
  for ( int i = nodeHistory.size () - 1; i >= 0; i-- ) {
    Node *node2 = nodeHistory.element ( i ).node;
    if ( isequalnode ( node2, 
                       modetable.modeData ( MODE ( node2 ) ),
		       newnode,
		       md ) ) 
      return true;
  }
  return false;
}

////////////////////////////////////////////////////////////////////////////////////////////

FArray<int> *substoldtonew, // the permutation corresponds to a substitution of variables.
                            // This array stores the substitution Original-->New
            *substnewtoold; // New-->Original
FArray<int> pernewtoold, // permutation of atoms. new-->old
            peroldtonew; // permutation of atoms. old-->new. old==-1 not assigned yet
int* iso_variables; // for each type the highest variable number used, used while
                    // building the isomorphic query, in the same way as during generation.
int iso_depth;
FArray<Node> node_cache; // the history of nodes in the equivalent query.
FArray<Node*> node_map; // each node in new query should occur in order as a path in the trie.
                 // We will use this property to remove the caches that we introduced
		 // temporarily.

Node *appendToQuery ( int newmode, int newdepth, Node *oldnode, int olddepth );

void printEquivalentHistory ( int newdepth ) {
  int i;
  for ( i = 0; i < newdepth-1; i++ ) {
    printNode ( node_map[i] );
    cout << ",";
  }
  printNode ( node_map[i] );
}

void makeSureThatArraysHaveSpace () {
  int *variables = nodeHistory.queryVariableCounts ();
  for ( int i = 0; i < typetable.size (); i++ ) {
    if ( variables[i] ) {
      for ( int l = substnewtoold[i].size (); l < variables[i]; l++ ) {
        substnewtoold[i].insertAt ( l, -1 );
        substoldtonew[i].insertAt ( l, -1 );
      }
      substoldtonew[key->params[0].number][0] = 0;
    }
    // init the complete arrays to -1?
  }
  
  int g;
  for ( g = pernewtoold.size (); g < nodeHistory.size (); g++ ) {
    pernewtoold.insertAt ( g, -1 );
    peroldtonew.insertAt ( g, -1 );
  }
  
  if ( node_cache.size () < nodeHistory.size () ) {
    for ( int k = node_cache.size (); k < nodeHistory.size (); k++ ) {
      Node dummynode;
      dummynode.params = new INTTYPE[32];
      node_cache.insertAt ( k, dummynode );
    }
    node_map.insertAt ( nodeHistory.size () - 1, NULL );
  }
}

// apply the key to the isomorphism structures.
// The key is the same for all queries, so this can be done one time.
void tree_prune_iso_init_key () {
  iso_variables = new int[typetable.size ()];
  for ( int i = 0; i < typetable.size (); i++ )
    iso_variables[i] = 0;
  iso_depth = 0;
  substnewtoold = new FArray<int> [typetable.size ()];
  substoldtonew = new FArray<int> [typetable.size ()];
  makeSureThatArraysHaveSpace ();

  Node *key = nodeHistory.element ( 0 ).node;
  node_map.insertAt ( 0, key );
  appendToQuery ( MODE ( key ), 0, key,0 );
}

// determines whether the atom `oldnode' can be added. Returns the mode number
// of the new mode for node oldnode, otherwise -1.
int canAppendToQuery ( Node *oldnode ) {
  ModePatternType m;
  init ( m );
  ModeData* md = modetable.modeData ( MODE ( oldnode ) );
  PredicateData *pd = predicatetable.predicateData ( md->predicate );

  // walk through parameters. For each variable, check whether it is already
  // mapped. If so: this is an input, otherwise: an output.
  for ( int i = 0; i < pd->size; i++ ) {
    if ( md->params[i].type != Constant && md->params[i].type != ChainConstant && substoldtonew[pd->params[i]][oldnode->params[i]] != -1 ) {
      // inuts are marked with a `one'.
      setOn ( m, i );
    }
  }

  int nr = modetable.searchEquivalentMode ( oldnode, m );
  return nr;
}

// determines whether the atom `oldnode' can be added with mode newmode (which is the first index in the
// modetable), which is part of a chain. Returns -1 if not possible, and the mode number otherwise.
// The returned mode number is the real mode (not the `firstnr' mode).
int canAppendToQueryChainWithMode ( Node *oldnode, ModeData *newmode, int newdepth ) {
  ModeData *oldmode = modetable.modeData ( MODE ( oldnode ) );
  ModeParamData *oldmpd = oldmode->params;
  int oldnrconst = oldmode->nrconst;
  ModeParamData *newmpd = newmode->params;
  int newnrconst = newmode->nrconst;
  PredicateData *pd = predicatetable.predicateData ( oldmode->predicate );
  int add = 0;
  
  if ( oldmode->predicate != newmode->predicate || oldnrconst != newnrconst )
    return -1;
  
  // walk through the parameters of the new mode.
  // stop as soon as a violation is found.
  for ( int i = 0; i < newmode->size; i++, oldmpd++, newmpd++ ) {
    if ( IS_VAR_PARAM ( oldmpd ) )
      if ( IS_VAR_PARAM ( newmpd ) )
	if ( newmpd->type == MinEx ) {
          if ( substoldtonew[pd->params[i]][oldnode->params[i]] != -1 )
	    return -1;
        }
	else {
	  if ( newmpd->type == Plus || newmpd->type == PlusEx ) {
            if ( substoldtonew[pd->params[i]][oldnode->params[i]] == -1 )
	      return -1;
	  }
	  else {
	    if ( newmpd->type == ChainPlus && 	     
	         substoldtonew[pd->params[i]][oldnode->params[i]] != node_cache[newdepth + newmpd->back].params[newmpd->backparamnr] )
	      return -1;
          }
	}
      else
        return -1;
    else
      if ( IS_VAR_PARAM ( newmpd ) )
        return -1;
      else
        if ( newmpd->type == ChainConstant ) {
	  if ( oldnode->params[i] != node_cache[newdepth + newmpd->back].params[newmpd->backparamnr] )
	    return -1;
	}
	else
	  if ( newmpd->type == Constant ) {
	    TypeData *td = typetable.typeData ( pd->params[i] );
	    add = add * td->allowedNrOfConstants () + ( oldnode->params[i] - td->nrforbidden );
	  }
	
  }
  return newmode->firstnr + add;
}

int canAppendToQueryChain ( Node *oldnode, int deepmodenr, int newdepth ) {
  FArray<ModeData*> &chainedmodes = modetable.chainedmodes ( deepmodenr );
  int val;

  for ( int i = 0; i < chainedmodes.size (); i++ ) {
    val = canAppendToQueryChainWithMode ( oldnode, chainedmodes[i], newdepth );
    if ( val != -1 )
      return val;
  }
  return -1;
}

// adds the oldnode to the current query being built. Makes sure the permutation is 
// filled in, and substitutions are performed.
// The new node is returned. It should be more efficient not to allocate
// this node each time anew (we could, for example, allocate an array of nodes
// for this purpose).
Node *appendToQuery ( int newmode, int newdepth, Node *oldnode, int olddepth ) {
  Node *newnode = &(node_cache[newdepth]);

  INTTYPE *oldparams = oldnode->params;
  INTTYPE *params = newnode->params;
  newnode->mode = newmode;

  // If no such mode, this atom is not appendable.
  // If a mode exist, we assign output variable names, and extend the substitution.

  // It seems a good idea to use a static node and a large array of parameters,
  // at this saves memory allocation. The returned node can however not be used
  // after a second call to this function. The variable names could be filled
  // in while searching the pattern, but we should take care of undoing in that case
  // if a pattern is not possible. Probably it's easier and just as efficient
  // to delay the determination of output variables.
  ModeData *oldmd = modetable.modeData ( MODE ( oldnode ) );
  ModeData *newmd = modetable.modeData ( newmode );
  PredicateData *pd = predicatetable.predicateData ( oldmd->predicate );
  int *pdparams = pd->params;
  
  for ( int i = 0; i < pd->size; i++ ) {
    switch ( newmd->params[i].type ) {
      case Plus: case ChainPlus: case PlusEx: params[i] = substoldtonew[pdparams[i]][oldparams[i]]; break;
      case Constant: case ChainConstant: params[i] = oldparams[i]; break;
      case MinEx: 
        // newmd->params[i].type == Min
        // A Plus becomes an output. As we introduce a variable, we have to do some bookkeeping.
        // Especially, the new variable maps to an old variable.
        params[i] = iso_variables[pdparams[i]];
        substoldtonew[pdparams[i]][oldparams[i]] = params[i];
        substnewtoold[pdparams[i]][params[i]] = oldparams[i];
        iso_variables[pdparams[i]]++;
    }
  }
  
  peroldtonew[olddepth] = newdepth;
  pernewtoold[newdepth] = olddepth;
  return newnode;
}

// removes the last node from the query.
void removeFromQuery ( Node *newnode, int newdepth ) {
  // undo the substitutions.
  ModeData *newmd = modetable.modeData ( MODE ( newnode ) );
  PredicateData *pd = predicatetable.predicateData ( newmd->predicate );
  int *pdparams = pd->params;
  INTTYPE *params = newnode->params;
  
  for ( int i = 0; i < pd->size; i++ ) {
    if ( newmd->params[i].type == MinEx ) {
      substoldtonew
        [pdparams[i]]
	[substnewtoold[pdparams[i]][params[i]]] 
	= -1;

      iso_variables[pdparams[i]]--;
    }
  }
  
  peroldtonew[pernewtoold[newdepth]] = -1;
}

int keepequivalent; // if found to be equivalent, should we maintain
                     // it to maintain completeness?
extern int infreqsubcount, equivalentcount;

int currentmaxdepth; // = nodehistory.size

// nextinchain is the mode which should be applied next in the chain.
// If the previous atom in the query does not have a next in the chain,
// nextinchain should be -1.
bool searchforequivalentquery ( int newdepth, int nextinchain ) {
  int newmodenr;
  Node *enode;
  
  for ( int i = 1; i < nodeHistory.size (); i++ )
  //for ( int i = nodeHistory.size () - 1; i >= 0; i-- )
    if ( peroldtonew[i] == -1 ) {
      if ( nextinchain == -1 )
        newmodenr = canAppendToQuery ( nodeHistory.element ( i ).node );
      else 
        newmodenr = canAppendToQueryChain ( nodeHistory.element ( i ).node, nextinchain, newdepth );
      if ( newmodenr != -1 ) {
        Node *newnode = appendToQuery ( newmodenr, newdepth, nodeHistory.element ( i ).node, i );
	//if ( nodehistory[i]->mode != newnode->mode ) {
	//cout << "FROM: " << endl;
	//printHistory ( nodehistory, treedepth + 1 );
	//cout << "TO: " << endl;
	//printHistory ( node_cache_history, newdepth + 1 );
	//}
        Node *parent = node_map[newdepth-1];
	// check frequent?
	if ( parent->infreqchildren && parent->infreqchildren->search ( newnode ) ) {
          removeFromQuery ( newnode, newdepth );
	  //cout << "[INFREQUENT PRUNED] ";
	    //printHistory ( nodehistory, treedepth + 1 );
	    //cout << "--> "; printHistory ( node_cache_history, newdepth + 1 );
	  infreqsubcount++;
	  return true;
	}
	// search in the frequent nodes.
	if ( parent == nodeHistory.element ( newdepth - 1 ).node ) {
	  // until now, we follow the same path as the original query.
	  if ( newdepth < nodeHistory.size () - 1 ) {
  	    TreeNodePointer *tnpr = &nodeHistory.element ( newdepth );
  	    TreeNodePointer *tnp;
	    if ( parent->freqchildren ) {
	      tnp = parent->freqchildren->search ( newnode );
	      if ( tnp && nodeCmp2b ( tnp, tnpr ) > 0 ) // is more "right in the tree!
	        tnp = NULL;
	    }
	    else
	      tnp = NULL;
	    if ( tnp ) {
	      node_map[newdepth] = tnp->node;
  	      if ( ( tnp->node == nodeHistory.element ( newdepth ).node || tnp->node->maxdepth >= currentmaxdepth ) && 
	           searchforequivalentquery ( newdepth + 1, modetable.chaindeepnr ( newmodenr ) ) ) {
	        removeFromQuery ( newnode, newdepth );
	        return true;
	      }	      
	    }
	  }
	  else {
	    // childlist has been artificially filled with the
	    // children constructed until now.
	    // remember that children are added in order!!!
	    TreeNodePointer *tnp;
	    if ( tnp = nodeHistory.last().childlist->search ( newnode ) ) {
	      removeFromQuery ( newnode, newdepth );
	      keepequivalent = tnp->node->support + 1;
	      equivalentcount++;
	      return true;
	    }
	    // have found ourselves
	  }
	}
	else {
	  TreeNodePointer *tnp;
	  if ( parent->freqchildren )
	    tnp = parent->freqchildren->search ( newnode );
	  else
	    tnp = NULL;
	  if ( tnp && newdepth == nodeHistory.size () - 1 ) {
	    removeFromQuery ( newnode, newdepth );
	    equivalentcount++;
	    return true;
	  }
	  else {
	    if ( tnp ) { 
	      node_map[newdepth] = tnp->node;
  	      if ( ( tnp->node == nodeHistory.element ( newdepth ).node || tnp->node->maxdepth >= currentmaxdepth ) && 
	           searchforequivalentquery ( newdepth + 1, modetable.chaindeepnr ( newmodenr ) ) ) {
	        removeFromQuery ( newnode, newdepth );
	        return true;
	      }
	    }
	  }
        }
	removeFromQuery ( newnode, newdepth );
      }
    }
  return false;
}

bool searchforequivalentquery () {
  // in principal, generate all permutations of this query.
  //cout << "---- C: ";
  //printHistory ( nodehistory, treedepth + 1 );
  keepequivalent = 0;
  if ( modetable.haschainsucc ( nodeHistory.last().node->mode ) )
    return false;
  
  if ( !nodeHistory.equivalentPossible ( nodeHistory.size () - 1 ) ) 
    return false;
  
  if ( peroldtonew.size () == 0 )
    tree_prune_iso_init_key ();
  
  makeSureThatArraysHaveSpace ();
  currentmaxdepth = nodeHistory.size ();
  
  bool b = searchforequivalentquery ( 1, key->chaindeepnr );
  /*if ( b )
    cout << " TRUE" << endl;
  else
    cout << " FALSE" << endl;*/
  return b;
}
