/***************************************************************************
                          typetable.cc  -  the dictionary of types
                             -------------------
    begin                : Tue Mar 21 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "output.h"
#include "typetable.h"

TypeTable::TypeTable ():
  actunroftypes ( 0 ),
  number_imm ( true ) {
}

TypeTable::~TypeTable () {
}

int TypeTable::addConstant ( char *constant_name, int typenr ) {
  TypeData *td = array[typenr];
  
  return addConstant ( constant_name, td );
}

int TypeTable::addConstant ( char *constant_name, char *type_name ) {
  TypeData *td = (TypeData*) typenames.lookup ( type_name );

  if ( !td ) {
    ERR << "Error: wrong type name (internal)" << endl;
    ERR_F
    return -1;
  }
  else
    return addConstant ( constant_name, td );
}

int TypeTable::addConstant ( char *constant_name, TypeData *td ) {
  if ( number_imm && constant_name[0] >= '0' && constant_name[0] <= '9' ) {
    ERR << "Error: can not use number as constant name" << endl;
    ERR_F
    return -1;
  }

  int val = td->constanttable.constantnametable->add ( constant_name, NULL );
  td->constnumbers++;
  return val;
}

int TypeTable::addType ( char *type_name ) {
  TypeData *td = new TypeData;
  td->constanttable.table = new NameTable;
  td->names = true;
  td->constnumbers = 0;
  td->nrforbidden = 0;
  td->oi = true;
  typenames.add ( type_name, td );
  actunroftypes++;
  array.append ( td );

  return actunroftypes - 1;
}

void TypeTable::getType ( char *type_name, int &typenr, int &nrofconstants ) {
  int ordernumber;
  TypeData *type = (TypeData*) typenames.lookup ( type_name, &ordernumber );
  if ( type ) {
    typenr = ordernumber;
    nrofconstants = type->constnumbers;
  }
  else {
    typenr = -1;
    nrofconstants = -1;
  }
}

// typenr should be filled in
void TypeTable::getConstant ( char *constant_name, int &typenr, int &constantnr ) {
  TypeData *type;

  if ( number_imm && constant_name[0] >= '0' && constant_name[0] <= '9' ) {
    typenr = -1; // unknown, no type... depend on what is later expected
    constantnr = atoi ( constant_name );
  }
  else {
    typenames.lookup ( typenr, (void**) &type );
    constantnr = type->constanttable.constantnametable->ordernumber ( constant_name );
  }
}


char *TypeTable::getConstantName ( int typenr, int constantnr ) {
  static char buffer[20];

  TypeData *type = array[typenr];

  if ( !type->constanttable.table ) {
    // no further info specified!
    sprintf ( buffer, "%i", constantnr );
    // no name, print number...

    return buffer;
  }
  return type->constanttable.constantnametable->lookup ( constantnr );
}

char *TypeTable::getTypeName ( int typenr ) {
  return typenames.lookup ( typenr );
}

void TypeTable::print () {
  cout << "+------+-----------------+" << endl
       << "| Type | Nr of constants |" << endl
       << "+------+-----------------+" << endl;
  for ( int i = 0; i < array.size (); i++ )
    cout << "| " << setw ( 4 ) << i << " |" << setw( 16 ) << array[i]->constnumbers << " |" << endl;
  cout << "+------+-----------------+" << endl;
}



