/***************************************************************************
                          typetable.h  - the dictionary of types 
                             -------------------
    begin                : Tue Mar 21 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef TYPETABLE_H
#define TYPETABLE_H
#include "nametable.h"
#include "floattable.h"
#include "farray.h"

struct TypeData {
   // constant names or floats belonging with the numbers, mixing
   // not allowed, determined by the first type added
  union {
    NameTable *constantnametable;
    FloatTable *floattable;
    void *table;
  } constanttable;
  bool names;
  bool oi;
  int constnumbers;
  FArray<bool> forbidden;
  int nrforbidden;
  inline int allowedNrOfConstants () const { return constnumbers - nrforbidden; }
};

class TypeTable {
  public:
    TypeTable ();
    ~TypeTable ();
    int addType ( char *type_name );
      // n upperbound of constants or floats, might be less when giving the list
    void getType ( char *type_name, int &typenr, int &nrofconstants );
    int getConstant ( int typenr, char *constant_name );
    void getConstant ( char *constant_name, int &typenr, int &constantnr );
    int addConstant ( char *constant_name, char *type_name );
    int addConstant ( char *constant_name, int typenr );
    void addForbid ( int typenr, int constantnr ) { 
      array[typenr]->forbidden.insertAt ( constantnr, true );
      array[typenr]->nrforbidden++;
    }
    char *getConstantName ( int typenr, int constantnr );
    char *getTypeName ( int typenr );
    void print ();
    void set_number_imm ( bool val ) { number_imm = val; }
    inline TypeData *typeData ( int i ) const { return array[i]; }
    inline int size () const { return array.size (); }
  private:
    FArray<TypeData*> array;
    int addConstant ( char *constant_name, TypeData *td );
    NameTable typenames;
    int actunroftypes;
    bool number_imm;
};

#endif






